""" Testing the radius cutoff utility
"""

from particula.util.radius_cutoff import cut_rad


def test_cuts():
    """ testing cuts:
            * test if starting radius is lower than mode
            * test if ending radius is higher than mod
            * test if lower radius is smaller than end radius
            * test if lower radius is smaller when cutoff is smaller
            * test if ending radius is larger when cutoff is larger
            * test if ending radius is larger when gsigma is higher
    """

    assert cut_rad(cutoff=.9999, gsigma=1.25, mode=1e-7)[0] <= 1e-7

    assert cut_rad(cutoff=.9999, gsigma=1.25, mode=1e-7)[1] >= 1e-7

    assert (
        cut_rad(cutoff=.9999, gsigma=1.25, mode=1e-7)[0]
        <=
        cut_rad(cutoff=.9999, gsigma=1.25, mode=1e-7)[1]
    )

    assert (
        cut_rad(cutoff=.9999, gsigma=1.25, mode=1e-7)[0]
        <=
        cut_rad(cutoff=.9990, gsigma=1.25, mode=1e-7)[0]
    )

    assert (
        cut_rad(cutoff=.9999, gsigma=1.25, mode=1e-7)[1]
        >=
        cut_rad(cutoff=.9990, gsigma=1.25, mode=1e-7)[1]
    )

    assert (
        cut_rad(cutoff=.9999, gsigma=1.25, mode=1e-7)[1]
        <=
        cut_rad(cutoff=.9999, gsigma=1.35, mode=1e-7)[1]
    )


def test_multi_cuts():
    """ test case for different modes
    """

    assert (
        cut_rad(cutoff=.9999, gsigma=1.25, mode=[1e-7, 1e-8])[0]
        <=
        cut_rad(cutoff=.9999, gsigma=1.25, mode=[1e-7, 1e-8])[1]
    )

    assert (
        cut_rad(cutoff=.9999, gsigma=1.25, mode=[1e-7, 1e-8])[0]
        <=
        cut_rad(cutoff=.9990, gsigma=1.25, mode=[1e-7, 1e-8])[0]
    )

    assert (
        cut_rad(cutoff=.9999, gsigma=1.25, mode=[1e-7, 1e-8])[1]
        >=
        cut_rad(cutoff=.9990, gsigma=1.25, mode=[1e-7, 1e-8])[1]
    )

    assert (
        cut_rad(cutoff=.9999, gsigma=1.25, mode=[1e-7, 1e-8])[1]
        <=
        cut_rad(cutoff=.9999, gsigma=1.25, mode=[1e-7, 1e-8])[1]
    )

    assert (
        cut_rad(cutoff=.9999, gsigma=1.25, mode=[1e-7, 1e-8])[1]
        ==
        cut_rad(cutoff=.9999, gsigma=1.25, mode=[1e-7])[1]
    )

    assert (
        cut_rad(cutoff=.9999, gsigma=1.25, mode=[1e-7, 1e-8])[1]
        >=
        cut_rad(cutoff=.9999, gsigma=1.25, mode=[1e-8])[1]
    )

    assert (
        cut_rad(cutoff=.9999, gsigma=1.25, mode=[1e-7, 1e-8])[0]
        ==
        cut_rad(cutoff=.9999, gsigma=1.25, mode=[1e-8])[0]
    )

    assert (
        cut_rad(cutoff=.9999, gsigma=1.25, mode=[1e-7, 1e-8])[0]
        <=
        cut_rad(cutoff=.9999, gsigma=1.25, mode=[1e-7])[0]
    )


def test_custom_radius():
    """ testing customizing the radius
    """
    assert cut_rad(
        cutoff=.9999, gsigma=1.25, mode=1e-7,
        force_radius_start=1e-8, force_radius_end=1e-6
    ) == (1e-8, 1e-6)
    assert cut_rad(
        cutoff=.9999, gsigma=1.25, mode=[1e-7, 2e-9],
        force_radius_start=1e-8, force_radius_end=1e-6
    ) == (1e-8, 1e-6)



""" calculate the species properties, saturation vapor pressures,
    latent heats, etc.

    add a more general class, with a dictionary that can be added from
    a file or inputs.
"""
# pylint: disable=all
import numpy as np
from particula import u
from particula.util.input_handling import (in_latent_heat, in_temperature,
                                           in_pressure, in_concentration,
                                           in_scalar
                                           )
from particula.constants import GAS_CONSTANT


class MaterialProperties:
    def __init__(self, species_properties):
        self.species_properties = species_properties

    def __getitem__(self, species):
        return self.species_properties[species]

    def saturation_pressure(self, temperature, species):
        temperature = in_temperature(temperature)
        try:
            if species == 'water':
                return self[species]['saturation_pressure'](temperature)
            else:
                return in_pressure(self[species]['saturation_pressure'])
        except KeyError:
            raise ValueError("Species not implemented")

    def latent_heat(self, temperature, species):
        temperature = in_temperature(temperature)
        try:
            if species == 'water':
                return in_latent_heat(
                        self[species]['latent_heat'](temperature)
                    )
            else:
                return in_latent_heat(self[species]['latent_heat'])
        except KeyError:
            raise ValueError("Species not implemented")


def water_buck_psat(temperature):
    """ Buck equation for water vapor pressure
        https://en.wikipedia.org/wiki/Arden_Buck_equation
    """
    temperature = in_temperature(temperature)
    temp = temperature.m_as("degC")

    return in_pressure(
        6.1115 * np.exp(
            (23.036-temp/333.7)*(temp/(279.82+temp))
        )*u.hPa * (temp < 0.0) + 6.1121 * np.exp(
            (18.678-temp/234.5)*(temp/(257.14+temp))
        )*u.hPa * (temp >= 0.0)
    )


def clausius_clapeyron(
            temperature,
            vapor_pressure,
            temperature_new,
            heat_vaporization
        ):
    """
    Calculates the vapor pressure of a substance at a given temperature
    using the Clausius-Clapeyron equation.

    Parameters
    ----------
        temperature (float): Temperature reference in Kelvin
        vapor_pressure (float): Vapor pressure reference in Pa
        temperature_new (float): Temperature new in Kelvin
        heat_vaporization (float): Heat of vaporization in J/kg

    Returns
    -------
        vapor_pressure_new: Vapor pressure in Pa
    """
    temperature = in_temperature(temperature)
    temperature_new = in_temperature(temperature_new)
    vapor_pressure = in_pressure(vapor_pressure)
    heat_vaporization = in_latent_heat(heat_vaporization)

    return vapor_pressure * np.exp(
        (-heat_vaporization / GAS_CONSTANT)
        * ((1 / temperature_new) - (1 / temperature))
    )


# maybe this should be loaded from a file, or for an option of the user to
# add their own species or load files with species properties
species_properties = {
    "generic": {
        "molecular_weight": 200.0 * u.g / u.mol,
        "surface_tension": 0.072 * u.N / u.m,
        "density": 1000.0 * u.kg / u.m ** 3,
        "vapor_radius": 1e-14 * u.Pa,
        "vapor_attachment": 1.0 * u.dimensionless,
        "saturation_pressure": 1.0 * u.Pa,
        "latent_heat": 1.0 * u.J / u.g,
        "heat_vaporization": 1.0 * u.J / u.kg,
        "kappa": 0.2 * u.dimensionless,
    },

    "water": {
        "molecular_weight": 18.01528 * u.g / u.mol,
        "surface_tension": 0.072 * u.N / u.m,
        "density": 1000.0 * u.kg / u.m ** 3,
        "vapor_radius": 1.6e-9 * u.m,
        "vapor_attachment": 1.0 * u.dimensionless,
        "saturation_pressure": water_buck_psat,
        "latent_heat": lambda T: (
                2500.8 - 2.36*T.m_as('degC')
                + 0.0016*T.m_as('degC')**2
                - 0.00006*T.m_as('degC')**3
            ) * u.J/u.g,
        "heat_vaporization": 2.257e6 * u.J / u.kg,
        "kappa": 0.2 * u.dimensionless,
    },

    "ammonium sulfate": {
        "molecular_weight": 132.14 * u.g / u.mol,
        "surface_tension": 0.072 * u.N / u.m,
        "density": 1770.0 * u.kg / u.m ** 3,
        "vapor_radius": 1.6e-9 * u.m,
        "vapor_attachment": 1.0 * u.dimensionless,
        "saturation_pressure": 1e-14 * u.Pa,
        "latent_heat": 0.0 * u.J/u.g,
        "heat_vaporization": 0.0 * u.J / u.kg,
        "kappa": 0.53 * u.dimensionless,
    }
}


def material_properties(property, species="water", temperature=298.15 * u.K):
    """Return the material properties for a given species.

    Parameters
    ----------
    property : str
        Property to return. Options are: 'all', 'molecular_weight',
        'surface_tension', 'density', 'vapor_radius', 'vapor_attachment',
        'kappa'.
    species : str
        Species for which to return the material properties.
    temperature : K
        Temperature of the material.

    Returns
    -------
    material_properties : value
        The material property for the given species.
    """

    if species in species_properties:
        material_props = MaterialProperties(species_properties)

    if property == 'all':
        return material_props[species]
    elif property in material_props[species]:
        if property == 'saturation_pressure':
            return material_props.saturation_pressure(temperature, species)
        elif property == 'latent_heat':
            return material_props.latent_heat(temperature, species)
        else:
            # no temperature dependence implemented
            return material_props[species][property]
    else:
        raise ValueError("Property not implemented")


def vapor_concentration(
        saturation_ratio,
        temperature=298.15 * u.K,
        species="water"
        ):
    """Convert saturation ratio to mass concentration at a given temperature.

    Parameters
    ----------
    sat_ratio : float
        saturation ratio.
    temperature : K
        Air temperature.
    species : str, optional
        Species for which to calculate the saturation vapor pressure.
        Default is "water".

    Returns
    -------
    concentration : concentration units
        Concentration vapor.

    TODO: check values for different species
    """
    saturation_ratio = in_scalar(saturation_ratio)
    # Calculate saturation vapor pressure
    saturation_pressure = material_properties(
            'saturation_pressure',
            species,
            temperature
        )

    return in_concentration(
        saturation_ratio
        * saturation_pressure
        / (GAS_CONSTANT * temperature)
        * material_properties('molecular_weight', species)
    )



"""Test the Stream class."""


import numpy as np
from particula.data.stream import Stream, StreamAveraged


def test_stream_initialization():
    """Test the initialization of the Stream class."""
    # Arrange
    header = ['header1', 'header2']
    data = np.array([1, 2, 3])
    time = np.array([1.0, 2.0, 3.0])
    files = ['file1', 'file2']

    # Act
    # stream = Stream(header=header)
    stream = Stream(header=header, data=data, time=time, files=files)

    # Assert
    assert stream.header == header
    assert np.array_equal(stream.data, data)
    assert np.array_equal(stream.time, time)
    assert stream.files == files


def test_stream_property_datetime64():
    """Test the datetime64 property of the Stream class."""
    # Arrange
    header = ['header1', 'header2']
    data = np.array([1, 2, 3])
    time = np.array([1.0, 2.0, 3.0])
    files = ['file1', 'file2']
    stream = Stream(header=header, data=data, time=time, files=files)

    # Act
    datetime64 = stream.datetime64

    # Assert
    assert np.array_equal(
        datetime64, np.array(['1970-01-01T00:00:01.000000000',
                              '1970-01-01T00:00:02.000000000',
                              '1970-01-01T00:00:03.000000000'],
                             dtype='datetime64[ns]'))


def test_stream_averaged_initialization():
    """Test the initialization of the StreamAveraged class."""
    # Arrange
    header = ['header1', 'header2']
    data = np.array([1, 2, 3])
    time = np.array([1.0, 2.0, 3.0])
    files = ['file1', 'file2']
    average_interval = 1.0
    start_time = 0.0
    stop_time = 2.0
    standard_deviation = np.array([0.1, 0.2, 0.3])

    # Act
    stream_averaged = StreamAveraged(
        header=header,
        data=data,
        time=time,
        files=files,
        average_interval=average_interval,
        start_time=start_time,
        stop_time=stop_time,
        standard_deviation=standard_deviation)

    # Assert
    assert stream_averaged.header == header
    assert np.array_equal(stream_averaged.data, data)
    assert np.array_equal(stream_averaged.time, time)
    assert stream_averaged.files == files
    assert stream_averaged.average_interval == average_interval
    assert stream_averaged.start_time == start_time
    assert stream_averaged.stop_time == stop_time
    assert np.array_equal(stream_averaged.standard_deviation,
                          standard_deviation)



"""File readers and loaders for datacula."""

from typing import List, Union, Tuple, Dict, Any, Optional
import warnings
import glob
import os
import pickle
import netCDF4 as nc
import numpy as np
import pandas as pd

from particula.util import convert
from particula.util.time_manage import time_str_to_epoch
from particula.data.lake import Lake

FILTER_WARNING_FRACTION = 0.5


def data_raw_loader(file_path: str) -> list:
    """
    Load raw data from a file at the specified file path and return it as a
    list of strings.

    Args:
        file_path (str): The file path of the file to read.

    Returns:
        list: The raw data read from the file as a list of strings.

    Examples:
        >>> data = data_raw_loader('my_file.txt')
        Loading data from: my_file.txt
        >>> print(data)
        ['line 1', 'line 2', 'line 3']
    """
    try:
        with open(file_path, 'r', encoding='utf8', errors='replace') as file:
            data = [line.rstrip() for line in file]
    except FileNotFoundError:
        print(f"File not found: {file_path}")
        data = []
    return data


def filter_list(data: List[str], char_counts: dict) -> List[str]:
    """
    A pass filter of rows from a list of strings.
    Each row must contain a specified number of characters to pass the filter.
    The number of characters to count is specified in the char_counts
    dictionary. The keys are the characters to count, and the values are the
    exact count required for each character in each row.

    Args:
    ----------
        data (List[str]): A list of strings to filter.
            A list of strings to filter.
        char_counts (dict): A dictionary of character counts to select by.
            The keys are the characters to count, and the values are the
            count required for each character.

    Returns:
    ----------
        List[str]: A new list of strings containing only the rows that meet the
        character count requirements.

    Raises:
    ----------
        UserWarning: If more than 90% of the rows are filtered out, and it
            includes the character(s) used in the filter.

    Examples:
    ----------
        >>> data = ['apple,banana,orange', 'pear,kiwi,plum',
                    'grapefruit,lemon']
        >>> char_counts = {',': 2}
        >>> filtered_data = filter_rows_by_count(data, char_counts)
        >>> print(filtered_data)
        ['apple,banana,orange', 'pear,kiwi,plum']
    """
    filtered_data = data
    for char, count in char_counts.items():
        if count > -1:
            filtered_data = [
                row for row in filtered_data if row.count(char) == count]
        if len(filtered_data) / len(data) < FILTER_WARNING_FRACTION:
            warnings.warn(
                f"More than {FILTER_WARNING_FRACTION} of the rows have " +
                f"been filtered out based on the character: {char}.")
    return filtered_data


def data_format_checks(data: List[str], data_checks: dict) -> List[str]:
    """
    Check if the data is in the correct format.

    Args:
        data (List[str]): A list of strings containing the raw data.
        data_checks (dict): Dictionary containing the format checks.

    Returns:
        List[str]: A list of strings containing the formatted data.

    Raises:
        TypeError: If data is not a list.

    Examples:
        >>> data = ['row 1', 'row 2', 'row 3']
        >>> data_checks = {
        ...     "characters": [0, 10],
        ...     "char_counts": {",": 2, "/": 0, ":": 0},
        ...     "skip_rows": 0,
        ...     "skip_end": 0
        ... }
        >>> formatted_data = data_format_checks(data, data_checks)
        >>> print(formatted_data)
        ['row 2']
    """
    if not isinstance(data, list):
        raise TypeError("data must be a list")
    length_initial = len(data)
    if data_checks.get('skip_rows', 0) > 0:
        data = data[data_checks['skip_rows']:]
    if data_checks.get('skip_end', 0) > 0:
        data = data[:-data_checks['skip_end']]
    if len(data_checks.get('characters', [])) == 1:
        # Filter out any rows with fewer than the specified number of
        # characters.
        data = [
            x for x in data
            if (
                len(x)
                > data_checks['characters'][0]
            )
        ]
    elif len(data_checks.get('characters', [])) == 2:
        # Filter out any rows with fewer than the minimum or more than the
        # maximum number of characters.
        data = [
            x for x in data
            if (
                data_checks['characters'][0]
                < len(x)
                < data_checks['characters'][1]
            )
        ]
    if len(data) / length_initial < FILTER_WARNING_FRACTION:
        warnings.warn(
            f"More than {FILTER_WARNING_FRACTION} rows are filtered based on "
            + f"{data_checks['characters']} or skip rows."
        )
    if 'char_counts' in data_checks:
        char_counts = data_checks.get('char_counts', {})
        data = filter_list(data, char_counts)
    if data := [x.strip() for x in data]:
        return data
    else:
        raise ValueError('No data left in file')


def parse_time_column(
    time_column: Union[int, List[int]],
    time_format: str,
    line: np.ndarray,
    date_offset: Optional[str] = None,
    seconds_shift: int = 0,
    timezone_identifier: str = 'UTC'
) -> float:
    """
    Parses the time column of a data line and returns it as a timestamp.

    Args:
    ----------
    time_column : Union[int, List[int]]
        The index or indices of the column(s) containing the time information.
    time_format : str
        The format of the time information, e.g. '%Y-%m-%d %H:%M:%S'.
    line : str
        The data line to parse.
    date_offset : Optional[str], default=None
        A fixed date offset to add to the timestamp in front.
    seconds_shift : int, default=0
        A number of seconds to add to the timestamp.

    Returns:
    -------
    float
        The timestamp corresponding to the time information in the data line,
        in seconds since the epoch.

    Raises:
    ------
    ValueError
        If an invalid time column or format is specified.
    """
    if time_format == 'epoch':
        return float(line[time_column]) + seconds_shift
    if date_offset:
        # if the time is in one column, and the date is fixed
        time_str = f"{date_offset} {line[time_column]}"
        return (
            time_str_to_epoch(time_str, time_format, timezone_identifier)
            + seconds_shift
        )
    if isinstance(time_column, int):
        return (
            time_str_to_epoch(
                line[time_column], time_format, timezone_identifier
            )
            + seconds_shift
        )
    if isinstance(time_column, list) and len(time_column) == 2:
        # if the time and date are in two column
        time_str = f"{line[time_column[0]]} {line[time_column[1]]}"
        return (
            time_str_to_epoch(time_str, time_format, timezone_identifier)
            + seconds_shift
        )
    raise ValueError(
        f"Invalid time column or format: {time_column}, {time_format}")


def sample_data(
    data: List[str],
    time_column: Union[int, List[int]],
    time_format: str,
    data_columns: List[int],
    delimiter: str,
    date_offset: Optional[str] = None,
    seconds_shift: int = 0,
    timezone_identifier: str = 'UTC'
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Samples the data to get the time and data streams.

    Args:
    -----------
    data : List[str]
        The input data in the form of a list of strings.
    time_column : int
        The index of the column that contains the time values.
    time_format : str
        The format string that specifies the time format.
    data_columns : List[int]
        The indices of the columns that contain the data values.
    delimiter : str
        The delimiter character used to separate columns in the input data.
    date_offset : str, optional
        A string that represents an offset in the date, in the format
        'days:hours:minutes:seconds'. Defaults to None.
    seconds_shift : int, optional
        An integer that represents a time shift in seconds. Defaults to 0.
    timezone_identifier : str, optional
        What timezone the data is in. Defaults to 'UTC'.

    Returns:
    --------
    Tuple[np.ndarray, np.ndarray]
        A tuple of two numpy arrays - epoch_time and data_array:
        - epoch_time : np.ndarray
            A 1-D numpy array of epoch times.
        - data_array : np.ndarray
            A 2-D numpy array of data values.

    Raises:
    -------
    ValueError:
        - If the data value is not in the correct format.
        - If no match for data value is found.
    """
    # flake8: noqa
    # pylint disable: too-many-arguments
    epoch_time = np.zeros(len(data))
    epoch_time = np.zeros(len(data))
    data_array = np.zeros((len(data), len(data_columns)))

    for i, line in enumerate(data):
        # split the line into an array
        line_array = np.array(line.split(delimiter))
        # split the line into an array
        line_array = np.array(line.split(delimiter))

        epoch_time[i] = parse_time_column(
            time_column=time_column,
            time_format=time_format,
            line=line_array,
            date_offset=date_offset,
            seconds_shift=seconds_shift,
            timezone_identifier=timezone_identifier
        )

        for j, col in enumerate(data_columns):
            value = line_array[col].strip() if col < len(line_array) else ''
            if value in ['', '.']:  # no data
                data_array[i, j] = np.nan
            elif value.count('�') > 0:
                data_array[i, j] = np.nan
            elif value[0].isnumeric():  # if the first character is a number
                data_array[i, j] = float(value)
            elif value[-1].isnumeric():
                data_array[i, j] = float(value)
            elif value[0] == '-':
                data_array[i, j] = float(value)
            elif value[0] == '+':
                data_array[i, j] = float(value)
            elif value[0] == '.':
                try:
                    data_array[i, j] = float(value)
                except ValueError as exc:
                    print(line_array)
                    raise ValueError(
                        f'Data is not a float: row {i}, col {j}, value {value}'
                    ) from exc

            elif value.isalpha():
                true_match = [
                    'ON', 'on', 'On', 'oN', '1', 'True', 'true',
                    'TRUE', 'tRUE', 't', 'T', 'Yes', 'yes', 'YES',
                    'yES', 'y', 'Y'
                ]
                false_match = [
                    'OFF', 'off', 'Off', 'oFF', '0',
                    'False', 'false', 'FALSE', 'fALSE', 'f',
                    'F', 'No', 'no', 'NO', 'nO', 'n', 'N'
                ]
                nan_match = [
                    'NaN', 'nan', 'Nan', 'nAN', 'NAN', 'NaN',
                    'nAn', 'naN', 'NA', 'Na', 'nA', 'na',
                    'N', 'n', '', 'aN', 'null', 'NULL', 'Null',
                    '-99999', '-9999', '.'
                ]
                if value in true_match:
                    data_array[i, j] = 1
                elif value in false_match:
                    data_array[i, j] = 0
                elif value in nan_match:
                    data_array[i, j] = np.nan
                else:
                    raise ValueError(
                        f'No match for data value: row {i}, \
                             col {j}, value {value}'
                    )

    return epoch_time, data_array


def general_data_formatter(
    data: list,
    data_checks: dict,
    data_column: list,
    time_column: Union[int, List[int]],
    time_format: str,
    delimiter: str = ',',
    header_row: int = 0,
    date_offset: Optional[str] = None,
    seconds_shift: int = 0,
    timezone_identifier: str = 'UTC'
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Formats and samples the data to get the time and data streams.

    Args:
    ----------
    data : list
        The list of strings containing the data.
    data_checks : dict
        A dictionary of data format checks to apply to the data.
    data_column : list
        The list of indices of the columns containing the data.
    time_column : Union[int, List[int]]
        The index or indices of the column(s) containing the time information.
    time_format : str
        The format of the time information, e.g. '%Y-%m-%d %H:%M:%S'.
    delimiter : str, default=','
        The delimiter used to separate columns in the data.
    date_offset : str, default=None
        A fixed date offset to add to the timestamp in front.
    seconds_shift : int, default=0
        A number of seconds to add to the timestamp.

    Returns:
    -------
    Tuple[np.array, np.array]
        A tuple containing two np.array objects: the first contains the
        epoch times, and the second contains the data.
    """

    # find str matching in header row and gets index
    if isinstance(data_column[0], str):
        data_header = data[header_row].split(delimiter)
        # Get data column indices
        data_column = [data_header.index(x)
                       for x in data_column]

    # Check the data format
    data = data_format_checks(data, data_checks)

    # Sample the data to get the epoch times and the data
    epoch_time, data_array = sample_data(
        data,
        time_column,
        time_format,
        data_column,
        delimiter,
        date_offset,
        seconds_shift,
        timezone_identifier
    )

    return epoch_time, data_array


def sizer_data_formatter(
    data: List[str],
    data_checks: Dict[str, Any],
    data_sizer_reader: Dict[str, str],
    time_column: Union[int, List[int]],
    time_format: str,
    delimiter: str = ',',
    header_row: int = 0,
    date_offset: Optional[str] = None,
    seconds_shift: int = 0,
    timezone_identifier: str = 'UTC'
) -> Tuple[np.ndarray, np.ndarray, list]:
    """
    Formats data from a particle sizer.

    Parameters
    ----------
    data : List[str]
        The data to be formatted.
    data_checks : Dict[str, Any]
        Dictionary specifying the formatting requirements for the data.
    data_sizer_reader : Dict[str, str]
        Dictionary containing information about the sizer data format.
    time_column : int
        The index of the time column in the data.
    time_format : str
        The format of the time information.
    delimiter : str, default=','
        The delimiter used in the data.
    date_offset : str, default=None
        The date offset to add to the timestamp.
    seconds_shift : int, default=0
        The number of seconds to add to the timestamp.
    timezone_identifier : str, default='UTC'
        The timezone identifier for the data.

    Returns
    -------
    Tuple[np.ndarray, List(str) np.ndarray, np.ndarray]
        A tuple containing the epoch time, the Dp header, and the data arrays.
    """

    # Get Dp range and columns
    data_header = data[header_row].split(delimiter)
    # check if start and end keywords are in the header
    if data_sizer_reader["Dp_start_keyword"] not in data_header:
        # rise error with snip of data header
        raise ValueError(
            f"Cannot find '{data_sizer_reader['Dp_start_keyword']}' in header"
            + f" {data_header[:20]}..."
        )
    if data_sizer_reader["Dp_end_keyword"] not in data_header:
        # rise error with snip of data header
        raise ValueError(
            f"Cannot find '{data_sizer_reader['Dp_end_keyword']}' in header"
            + f" {data_header[:20]}..."
        )
    dp_range = [
        data_header.index(data_sizer_reader["Dp_start_keyword"]),
        data_header.index(data_sizer_reader["Dp_end_keyword"])
    ]
    dp_columns = list(range(dp_range[0], dp_range[1] + 1))  # +1 to include end
    header = [data_header[i] for i in dp_columns]
    # change from np.array

    # Format data
    data = data_format_checks(data, data_checks)

    # Get data arrays
    epoch_time, data_2d = sample_data(
        data,
        time_column,
        time_format,
        dp_columns,
        delimiter,
        date_offset,
        seconds_shift=seconds_shift,
        timezone_identifier=timezone_identifier
    )

    if "convert_scale_from" in data_sizer_reader:
        if data_sizer_reader["convert_scale_from"] == "dw":
            for i in range(len(epoch_time)):
                data_2d[i, :] = convert.convert_sizer_dn(
                    diameter=np.array(header).astype(float),
                    dn_dlogdp=data_2d[i, :],
                    inverse=True
                )

    return epoch_time, data_2d, header


def non_standard_date_location(
    data: list,
    date_location: dict
) -> str:
    """
    Extracts the date from a non-standard location in the data.

    Args:
    ----------
    data : list
        A list of strings representing the data.
    date_location : dict
        A dictionary specifying the method for extracting the date from the
        data.
        Supported methods include:
            - 'file_header_block': The date is located in the file header
                block, and its position is specified by the 'row',
                'delimiter', and 'index' keys.

    Returns:
    -------
    str
        The date extracted from the specified location in the data.

    Raises:
    ------
    ValueError
        If an unsupported or invalid method is specified in date_location.
    """
    if date_location['method'] != 'file_header_block':
        raise ValueError('Invalid date location method specified')

    row_index = date_location['row']
    delimiter = date_location['delimiter']
    index = date_location['index']
    return data[row_index].split(delimiter)[index].strip()


def get_files_in_folder_with_size(
    path: str,
    subfolder: str,
    filename_regex: str,
    min_size: int = 10,
) -> Tuple[List[str], List[str], List[int]]:
    """
    Returns a list of files in the specified folder and subfolder that
    match the given filename pattern and have a size greater than the
    specified minimum size.

    Args:
    ----------
    path : str
        The path to the parent folder.
    subfolder : str
        The name of the subfolder containing the files.
    filename_regex : str
        A regular expression pattern for matching the filenames.
    min_size : int, optional
        The minimum file size in bytes (default is 10).

    Returns:
    -------
    Tuple[List[str], List[str], List[int]]
        A tuple containing three lists:
        - The filenames that match the pattern and size criteria
        - The full paths to the files
        - The file sizes in bytes
    """
    search_path = os.path.join(path, subfolder)

    if not os.path.isdir(search_path):
        raise ValueError(f"{search_path} is not a directory")

    file_list = glob.glob(os.path.join(search_path, filename_regex))

    # filter the files by size
    full_path = [
        file for file in file_list
        if os.path.getsize(os.path.join(search_path, file)) > min_size
    ]

    # get the file names only
    file_list = [os.path.split(path)[-1]
                 for path in full_path]
    file_size_in_bytes = [os.path.getsize(path) for path in full_path]

    return file_list, full_path, file_size_in_bytes


def save_lake(
        path: str,
        lake: Lake,
        sufix_name: Optional[str] = None):
    """
    Save lake object as a pickle file.

    Parameters
    ----------
    data_lake : DataLake
        DataLake object to be saved.
    path : str
        Path to save pickle file.
    sufix_name : str, optional
        Suffix to add to pickle file name. The default is None.
    """
    print('Saving lake...')
    # create output folder if it does not exist
    output_folder = os.path.join(path, 'output')
    os.makedirs(output_folder, exist_ok=True)

    # add suffix to file name if present
    file_name = f'lake_{sufix_name}.pk' if sufix_name is not None else 'lake.pk'
    # path to save pickle file
    file_path = os.path.join(output_folder, file_name)

    # save datalake
    with open(file_path, 'wb') as file:
        pickle.dump(lake, file)
    print('Lake saved')


def load_lake(path: str, sufix_name: Optional[str] = None) -> object:
    """
    Load datalake object from a pickle file.

    Parameters
    ----------
    path : str
        Path to load pickle file.

    Returns
    -------
    data_lake : DataLake
        Loaded DataLake object.
    """
    file_name = f'lake_{sufix_name}.pk' if sufix_name is not None else 'lake.pk'
    # path to load pickle file
    file_path = os.path.join(path, 'output', file_name)

    # load datalake
    with open(file_path, 'rb') as file:
        lake = pickle.load(file)

    return lake


# pylint: disable-all
def netcdf_get_epoch_time(
        file_path: str,
        settings: dict
) -> np.ndarray:
    """
    Given a netCDF file path and settings, returns an array of epoch times in
    seconds as a float.

    Currently only uses ARM 1.2 netCDF files (base_time + time_offset)

    Args:
    ----------
        file_path (str): The path to the netCDF file.
        settings (dict): A dictionary containing settings for the instrument.

    Returns:
    -------
        np.ndarray: An array of epoch times, in seconds as a float.
    """
    nc_file = nc.Dataset(file_path)  # type: ignore

    epoch_time = np.zeros(nc_file.dimensions['time'].size)

    for time_col in settings['time_column']:
        epoch_time += nc_file.variables.get(time_col)[:]
    epoch_time = np.array(epoch_time.astype(float))
    nc_file.close()

    return epoch_time


# pylint: disable-all
def netcdf_data_1d_load(
        file_path: str,
        settings: dict
) -> Tuple[np.ndarray, list, np.ndarray]:
    """
    Given a netCDF file path and settings, returns a tuple containing the
    epoch time, header, and data as a numpy array. We do apply the mask to the
    data, and fill the masked values with nan.

    Args:
    ----------
        file_path (str): The path to the netCDF file.
        settings (dict): A dictionary containing settings for the instrument.

    Returns:
    -------
        Tuple[np.ndarray, list, np.ndarray]: A tuple containing the epoch time,
        header, and data as a numpy array.

    Errors:
    ------
        KeyError: If the settings dictionary does not contain 'data_1d'.
    """
    # check if data_1d is in the settings dic
    if 'data_1d' not in settings['netcdf_reader']:
        raise KeyError("data_1d not in settings['netcdf_reader']")

    # get header
    header_1d = settings['netcdf_reader']['header_1d']

    nc_file = nc.Dataset(file_path)  # type: ignore
    # get epoch time
    epoch_time = netcdf_get_epoch_time(file_path, settings)

    # empty array to store data
    data_1d = np.zeros(
        (len(settings['netcdf_reader']['data_1d']),
         nc_file.dimensions['time'].size)
    )
    # select and fill masked array with nan
    for i, data_col in enumerate(settings['netcdf_reader']['data_1d']):
        try:
            data = nc_file.variables.get(data_col)[:]
            data_1d[i, :] = np.ma.filled(data.astype(float), np.nan)
        except (TypeError, KeyError):
            data_1d[i, :] = np.nan
            warnings.warn(data_col + " not found in the netCDF file")
    nc_file.close()

    # check data shape, transpose if necessary so that time is last dimension
    data_1d = convert.data_shape_check(
        time=epoch_time,
        data=data_1d,
        header=header_1d)

    return epoch_time, header_1d, data_1d


# pylint: disable-all
def netcdf_data_2d_load(
        file_path: str,
        settings: dict
) -> Tuple[np.ndarray, list, np.ndarray]:
    """
    Given a netCDF file path and settings, returns a tuple containing the
    epoch time, header, and data as a numpy array. We do apply the mask to the
    data, and fill the masked values with nan.

    Args:
    ----------
        file_path (str): The path to the netCDF file.
        settings (dict): A dictionary containing settings for the instrument.

    Returns:
    -------
        Tuple[np.ndarray, list, np.ndarray]: A tuple containing the epoch time,
        header, and data as a numpy array.

    Errors:
    ------
        KeyError: If the settings dictionary does not contain 'data_2d'.
    """
    # check if data_1d is in the settings dic
    if 'data_2d' not in settings['netcdf_reader']:
        raise KeyError("data_2d not in settings['netcdf_reader']")

    # get epoch time
    epoch_time = netcdf_get_epoch_time(file_path, settings)
    # load netcdf file
    nc_file = nc.Dataset(file_path)  # type: ignore

    # select data_2d
    data_2d = nc_file.variables.get(settings['netcdf_reader']['data_2d'])[:]
    # convert masked array to numpy array
    data_2d = np.ma.filled(data_2d.astype(float), np.nan)
    # get header
    header_2d = nc_file.variables.get(
        settings['netcdf_reader']['header_2d']
    )[:]
    nc_file.close()

    # convert header to list of strings
    header_2d = [str(item) for item in header_2d.tolist()]

    # check data shape, transpose if necessary so that time is last dimension
    data_2d = convert.data_shape_check(
        time=epoch_time,
        data=data_2d,
        header=header_2d)

    return epoch_time, header_2d, data_2d


def netcdf_info_print(file_path, file_return=False):
    """
    Prints information about a netCDF file. Useful for generating settings
    dictionaries.

    Args:
    ----------
        file_path (str): The path to the netCDF file.
        file_return (bool): If True, returns the netCDF file object.
            Defaults to False.

    Returns:
    -------
        nc_file (netCDF4.Dataset): The netCDF file object.
    """

    nc_file = nc.Dataset(file_path)  # type: ignore
    print("Dimensions:")
    for dim in nc_file.dimensions:
        print(dim, len(nc_file.dimensions[dim]))
    print("\nVariables:")
    for var in nc_file.variables:
        print(var,
              nc_file.variables[var].shape,
              nc_file.variables[var].dtype)
    print("\nHeaders:")
    for attr in nc_file.ncattrs():
        print(attr, "=", getattr(nc_file, attr))

    if file_return:
        return nc_file
    nc_file.close()
    return None



""" test kelvin correct """

from particula import u

from particula.util.kelvin_correction import kelvin_radius, kelvin_term


def test_kelvin_radius():
    """ test: parcelona.kelvin_radius """
    assert kelvin_radius(
        surface_tension=0.072 * u.N/u.m,
        molecular_weight=0.01815 * u.kg/u.mol,
        density=1000 * u.kg/u.m**3,
        temperature=300
    ).u == u.m


def test_kelvin_term():
    """ test: parcelona.kelvin_term """
    assert kelvin_term(
        surface_tension=0.072 * u.N/u.m,
        molecular_weight=0.01815 * u.kg/u.mol,
        density=1000 * u.kg/u.m**3,
        temperature=300,
        radius=1 * u.m
    ).u == u.dimensionless



"""interface to import data to a data stream"""


from typing import Optional
import os
import numpy as np
from particula.data import loader, merger
from particula.data.stream import Stream
from particula.data.lake import Lake
from particula.util import convert


def get_new_files(
        path: str,
        import_settings: dict,
        loaded_list: Optional[list] = None,
) -> tuple:
    """
    Scan a directory for new files based on import settings and stream status.

    This function looks for files in a specified path using import settings.
    It compares the new list of files with a pre-loaded list in the stream
    object to determine which files are new. The comparison is made based on
    file names and sizes. It returns a tuple with the paths of new files, a
    boolean indicating if this was the first pass, and a list of file
    information for new files.

    Args:
    ----------
    path : str
        The top-level directory path to scan for files.
    import_settings : dict
        A dictionary with 'relative_data_folder', 'filename_regex',
        and 'MIN_SIZE_BYTES' as keys
        used to specify the subfolder path and the regex pattern for filtering
        file names. It should also include 'min_size' key to specify the
        minimum size of the files to be considered.
    loaded_list : list of lists
        A list of lists with file names and sizes that have already been
        loaded. The default is None. If None, it will be assumed that no
        files have been loaded.

    Returns:
    -------
    tuple of (list, bool, list)
        A tuple containing a list of full paths of new files, a boolean
        indicating if no previous files were loaded (True if it's the first
        pass), and a list of lists with new file names and sizes.

    Returns:
    Raises:
    ------
    YourErrorType
        Explanation of when and why your error is raised and what it means.
    """
    # Validate the path and settings
    if not os.path.isdir(path):
        raise ValueError(
            f"The provided path does not exist or is not a directory: {path}")
    # Settings Verification
    required_keys = [
        'relative_data_folder',
        'filename_regex',
        'MIN_SIZE_BYTES']
    if any(key not in import_settings for key in required_keys):
        raise KeyError(
            f"import_settings must contain the following keys: {required_keys}"
        )
    # Loaded files verification
    if loaded_list is not None:
        if not all(isinstance(item, list) for item in loaded_list):
            raise TypeError(
                f"loaded_list must be a list of lists. It is currently a \
                list of {type(loaded_list[0])}"
            )
        if any(len(item) != 2 for item in loaded_list):
            raise ValueError(
                f"loaded_list must be a list of lists with 2 items. It is \
                currently a list of lists with {len(loaded_list[0])} items"
            )

    # import data based on the settings what type
    file_names, full_paths, file_sizes = loader.get_files_in_folder_with_size(
        path=path,
        subfolder=import_settings['relative_data_folder'],
        filename_regex=import_settings['filename_regex'],
        min_size=import_settings['MIN_SIZE_BYTES'])
    # combined the file names and sizes into a list of lists
    file_info = [[name, size]
                 for name, size in zip(file_names, file_sizes)]
    # check and compare the previous file list with the new list.
    if not loaded_list:
        first_pass = True
    else:
        first_pass = False
        # check if the files are the same
        new_full_paths = []
        new_file_info = []
        for i, comparison_list in enumerate(file_info):
            if comparison_list not in loaded_list:
                # keep the new files
                new_full_paths.append(full_paths[i])
                new_file_info.append(comparison_list)
        full_paths = new_full_paths  # replace to return only new files
        file_info = new_file_info
    return full_paths, first_pass, file_info


def load_files_interface(
        path: str,
        settings: dict,
        stream: Optional[Stream] = None,
) -> Stream:
    """
    Load files into a stream object based on settings.

    Args:
    ----------
    path : str
        The top-level directory path to scan for folders of data.
    folder_settings : dict
        A dictionary with keys corresponding to the stream names and values
        corresponding to the settings for each stream. The settings can
        be generated using the settings_generator function.
    stream : Stream, optional
        An instance of Stream class to be updated with loaded data. Defaults
        to a new Stream object.

    Returns:
    -------
    Stream
        The Stream object updated with the loaded data.
    """
    if stream is None:
        stream = Stream(
            header=[],
            data=np.array([]),
            time=np.array([]),
            files=[])
    # get the files to load
    full_paths, first_pass, file_info = get_new_files(
        path=path,
        import_settings=settings,
        loaded_list=stream.files
    )

    # load the data type
    for file_i, file_path in enumerate(full_paths):
        print('  Loading file:', file_info[file_i][0])

        if settings['data_loading_function'] == 'general_1d_load':
            stream = get_1d_stream(
                file_path=file_path,
                first_pass=first_pass,
                settings=settings,
                stream=stream
            )

        elif settings['data_loading_function'] == 'general_2d_load':
            stream = get_2d_stream(
                file_path=file_path,
                first_pass=first_pass,
                settings=settings,
                stream=stream
            )

        # elif (self.settings[key]['data_loading_function'] ==
        #         'netcdf_load'):
        #     self.initialise_netcdf_stream(key, path, first_pass)
        #     first_pass = False
        else:
            raise ValueError('Data loading function not recognized',
                             settings['data_loading_function'])

        stream.files.append(file_info[file_i])  # add file info as loaded
        first_pass = False  # set first pass to false after first file
    return stream


def load_folders_interface(
        path: str,
        folder_settings: dict,
        lake: Optional[Lake] = None,
) -> Lake:
    """
    Load files into a lake object based on settings.

    Args:
    ----------
    path : str
        The top-level directory path to scan for folders of data.
    folder_settings : dict
        A dictionary with keys corresponding to the stream names and values
        corresponding to the settings for each stream. The settings can
        be generated using the settings_generator function.
    lake : Lake, optional
        An instance of Lake class to be updated with loaded data. Defaults
        to a new Lake object.

    Returns:
    -------
    Lake
        The Lake object updated with the loaded data streams.
    """
    if lake is None:
        lake = Lake()

    # loop through the folders
    for key, file_settings in folder_settings.items():
        stream = lake[key] if key in lake.streams.keys() else None
        print('Folder Settings:', key)
        # call the load files interface
        lake[key] = load_files_interface(
            path=path,
            settings=file_settings,
            stream=stream
        )
    return lake


def get_1d_stream(
    file_path: str,
    settings: dict,
    first_pass: bool = True,
    stream: Optional[Stream] = None,
) -> Stream:
    """
    Loads and formats a 1D data stream from a file and initializes or updates
    a Stream object.

    Args:
    ----------
    file_path : str
        The path of the file to load data from.
    first_pass : bool
        Whether this is the first time data is being loaded. If True, the
        stream is initialized.
        If False, raises an error as only one file can be loaded.
    settings : dict
        A dictionary containing data formatting settings such as data checks,
        column names,
        time format, delimiter, and timezone information.
    stream : Stream, optional
        An instance of Stream class to be updated with loaded data. Defaults
        to a new Stream object.

    Returns:
    -------
    Stream
        The Stream object updated with the loaded data and corresponding time
        information.

    Raises:
    ------
    ValueError
        If `first_pass` is False, indicating data has already been loaded.
    TypeError
        If `settings` is not a dictionary.
    FileNotFoundError
        If the file specified by `file_path` does not exist.
    KeyError
        If any required keys are missing in the `settings` dictionary.
    """
    if stream is None:
        stream = Stream(
            header=[],
            data=np.array([]),
            time=np.array([]),
            files=[])
    # Input validation, should it be abstracted?
    if not isinstance(settings, dict):
        raise TypeError("The setting parameters must be in a dictionary.")

    required_keys = ['data_checks', 'data_column', 'time_column',
                     'time_format', 'delimiter', 'time_shift_seconds',
                     'timezone_identifier', 'data_header', 'header_row']
    if any(key not in settings for key in required_keys):
        missing_key = [key for key in required_keys if key not in settings]
        raise KeyError(
            f"The settings dictionary is missing required keys: {missing_key}")

    if not os.path.isfile(file_path):
        raise FileNotFoundError(f"The file path specified does not exist: \
                                {file_path}")

    if not isinstance(first_pass, bool):
        raise TypeError("The first_pass parameter must be a boolean.")

    # should should consolidate and abstract this
    data = loader.data_raw_loader(file_path=file_path)

    if 'date_location' in settings:
        date_offset = loader.non_standard_date_location(
            data=data,
            date_location=settings['date_location']
        )
    else:
        date_offset = None

    epoch_time, data = loader.general_data_formatter(
        data=data,
        data_checks=settings['data_checks'],
        data_column=settings['data_column'],
        time_column=settings['time_column'],
        time_format=settings['time_format'],
        delimiter=settings['delimiter'],
        header_row=settings['header_row'],
        date_offset=date_offset,
        seconds_shift=settings['time_shift_seconds'],
        timezone_identifier=settings['timezone_identifier']
    )

    # check data shape
    data = convert.data_shape_check(
        time=epoch_time,
        data=data,
        header=settings['data_header'])
    if first_pass:
        stream.header = settings['data_header']
        stream.data = data
        stream.time = epoch_time
    else:
        stream = merger.stream_add_data(
            stream=stream,
            time_new=epoch_time,
            data_new=data,
            header_check=True,
            header_new=settings['data_header']
        )
    return stream


def get_2d_stream(
    file_path: str,
    settings: dict,
    first_pass: bool = True,
    stream: Optional[Stream] = None,
) -> Stream:
    """
    Initializes a 2D stream using the settings in the DataLake object.

    Args:
    ----------
        key (str): The key of the stream to initialise.
        path (str): The path of the file to load data from.
        first_pass (bool): Whether this is the first time loading data.

    Returns:
    ----------
        None.
    """
    if stream is None:
        stream = Stream(
            header=[],
            data=np.array([]),
            time=np.array([]),
            files=[])
    # Input validation
    if not isinstance(settings, dict):
        raise TypeError("The setting parameters must be in a dictionary.")

    required_keys = ['data_checks', 'time_column',
                     'time_format', 'delimiter', 'time_shift_seconds',
                     'timezone_identifier', 'data_sizer_reader',
                     'header_row']
    if any(key not in settings for key in required_keys):
        missing_key = [key for key in required_keys if key not in settings]
        raise KeyError(
            f"The settings dictionary is missing required keys: {missing_key}")

    if not os.path.isfile(file_path):
        raise FileNotFoundError(f"The file path specified does not exist: \
                                {file_path}")

    if not isinstance(first_pass, bool):
        raise TypeError("The first_pass parameter must be a boolean.")

    # should should consolidate and abstract this
    data = loader.data_raw_loader(file_path=file_path)
    if 'date_location' in settings:
        date_offset = loader.non_standard_date_location(
            data=data,
            date_location=settings['date_location']
        )
    else:
        date_offset = None

    epoch_time, data, header = loader.sizer_data_formatter(
        data=data,
        data_checks=settings['data_checks'],
        data_sizer_reader=settings['data_sizer_reader'],
        time_column=settings['time_column'],
        time_format=settings['time_format'],
        delimiter=settings['delimiter'],
        header_row=settings['header_row'],
        date_offset=date_offset,
        seconds_shift=settings['time_shift_seconds'],
        timezone_identifier=settings['timezone_identifier']
    )

    # check data shape
    data = convert.data_shape_check(
        time=epoch_time,
        data=data,
        header=header)
    if first_pass:
        stream.header = header
        stream.data = data
        stream.time = epoch_time
    else:
        stream = merger.stream_add_data(
            stream=stream,
            time_new=epoch_time,
            data_new=data,
            header_check=True,
            header_new=header
        )
    return stream


# def initialise_netcdf_stream(
#     self,
#     key: str,
#     path: str,
#     first_pass: bool
# ) -> None:
#     """
#     Initialise a netcdf stream using the settings in the DataLake
#     object. This can load either 1D or 2D data, as specified in the
#     settings.
#     Args:
#     ----------
#         key (str): The key of the stream to initialise.
#         path (str): The path of the file to load data from.
#         first_pass (bool): Whether this is the first time loading data.

#     Returns:
#     ----------
#         None.
#     """
#     # ValueKey error if netcdf_reader not in settings
#     if 'netcdf_reader' not in self.settings[key]:
#         raise ValueError('netcdf_reader not in settings')

#     # Load the data 1d data
#     if 'data_1d' in self.settings[key]['netcdf_reader']:
#         epoch_time, header_1d, data_1d = loader.netcdf_data_1d_load(
#             file_path=path,
#             settings=self.settings[key])

#         if first_pass:  # create the stream
#             self.streams[
#                 self.settings[key]['data_stream_name'][0]
#             ] = stream(
#                 header_list=header_1d,
#                 average_times=[600],
#                 average_base=self.settings[key]['base_interval_sec']
#             )

#         self.streams[
#             self.settings[key]['data_stream_name'][0]
#         ].add_data(
#             time_stream=epoch_time,
#             data_stream=data_1d,
#         )

#     if 'data_2d' in self.settings[key]['netcdf_reader']:
#         epoch_time, header_2d, data_2d = loader.netcdf_data_2d_load(
#             file_path=path,
#             settings=self.settings[key])

#         if first_pass:  # create the stream
#             self.streams[
#                 self.settings[key]['data_stream_name'][1]
#             ] = stream(
#                 header_list=header_2d,
#                 average_times=[600],
#                 average_base=self.settings[key]['base_interval_sec']
#             )

#         self.streams[
#             self.settings[key]['data_stream_name'][1]
#         ].add_data(
#             time_stream=epoch_time,
#             data_stream=data_2d,
#             header_check=True,
#             header=header_2d
#         )



"""Test the integration module."""
import torch
from particula.lagrangian import integration


def test_integration_leapfrog():
    """Test leapfrog integration."""
    position = torch.tensor([0.0, 0.0, 0.0])
    velocity = torch.tensor([1.0, 0.0, 0.0])
    force = torch.tensor([0.0, 1.0, 0.0])
    mass = torch.tensor([1.0, 1.0, 1.0])
    time_step = 1.0
    expected_position = torch.tensor([1.0, 0.5, 0.0])
    expected_velocity = torch.tensor([1.0, 1.0, 0.0])
    result_position, result_velocity = integration.leapfrog(
        position, velocity, force, mass, time_step)
    assert torch.allclose(result_position, expected_position)
    assert torch.allclose(result_velocity, expected_velocity)



""" test the simple solver
"""

import pytest
import numpy as np
from particula import particle, rates
from particula.util import simple_solver

simple_dic_kwargs = {
    "mode": 200e-9,  # 200 nm median
    "nbins": 1000,  # 1000 bins
    "nparticles": 1e6,  # 1e4 #
    "volume": 1e-6,  # per 1e-6 m^3 (or 1 cc)
    "gsigma": 1.2,  # relatively narrow
    "cutoff": .99999,  # let's take it all lol
}

particle_dist = particle.Particle(**simple_dic_kwargs)

coag_kern = rates.Rates(particle=particle_dist).particle_coagulation

Solver = simple_solver.SimpleSolver(
    distribution=particle_dist.particle_distribution(),
    radius=particle_dist.particle_radius,
    kernel=coag_kern,
    tspan=np.linspace(0, 100, 100),
)

FinerSolver = simple_solver.SimpleSolver(
    distribution=particle_dist.particle_distribution(),
    radius=particle_dist.particle_radius,
    kernel=coag_kern,
    tspan=np.linspace(0, 100, 1000),
)

radius = particle_dist.particle_radius.m
solution = Solver.solution()
fine_sols = FinerSolver.solution()


def test_dims():
    """ test the dimensions
    """

    assert (
        Solver.prep_inputs()[0].shape ==
        particle_dist.particle_distribution().m.shape
    )
    assert (
        Solver.prep_inputs()[1].shape ==
        particle_dist.particle_radius.m.shape
    )
    assert (
        Solver.prep_inputs()[2].shape ==
        coag_kern.m.shape
    )
    assert (
        Solver.prep_inputs()[3].shape ==
        np.linspace(0, 100, 100).shape
    )


def test_solution():
    """ test the solution
    """

    assert solution.m.shape == (100, 1000)
    assert solution.u == particle_dist.particle_distribution().u
    assert fine_sols.m.shape == (1000, 1000)
    assert (solution.m[0, :] == fine_sols.m[0, :]).all()


def test_conservation():
    """ test the conservation
    """

    assert (
        np.trapz((solution.m[0, :]-solution.m[-1, :])*radius**3, radius) /
        np.trapz((solution.m[0, :])*radius**3, radius)
        ==
        pytest.approx(0, abs=1e-3)
    )

    assert (
        np.trapz((fine_sols.m[0, :]-fine_sols.m[-1, :])*radius**3, radius) /
        np.trapz((fine_sols.m[0, :])*radius**3, radius)
        ==
        pytest.approx(0, abs=1e-3)
    )



""" testing the distribution discretization
"""

import numpy as np
import pytest
from particula.util.distribution_discretization import discretize


def test_discretize():
    """ testing the discretization utility
    """

    spans = np.linspace(1, 1000, 1000)
    sigma = 1.25
    modes = 100

    assert discretize(
        interval=spans, disttype="lognormal", gsigma=sigma, mode=modes
    ).size == spans.size

    assert np.trapz(discretize(
        interval=spans, disttype="lognormal", gsigma=sigma, mode=modes
    ), spans) == pytest.approx(1, rel=1e-5)

    with pytest.raises(ValueError):
        discretize(
            interval=spans, disttype="linear", gsigma=sigma, mode=modes,
        )


def test_multi_discretize():
    """ testing different modes
    """

    spans = np.linspace(1, 1000, 1000)
    sigma = 1.25
    modes = [100, 200]

    assert discretize(
        interval=spans, disttype="lognormal", gsigma=sigma, mode=modes
    ).size == spans.size

    assert np.trapz(discretize(
        interval=spans, disttype="lognormal", gsigma=sigma, mode=modes
    ), spans) == pytest.approx(1, rel=1e-5)

    with pytest.raises(ValueError):
        discretize(
            interval=spans, disttype="linear", gsigma=sigma, mode=modes,
        )



""" test the diffusive knudsen number calculation
"""

import pytest
from particula import u
from particula.util.diffusive_knudsen import diff_knu


def test_diffusive_knudsen():
    """ test the diffusive knudsen number calculation

        What is the diffusive knudsen number?
        What does it mean?
        Why is calculated like this?
        How can we test it properly?
        Any rules of thumb here?
    """

    assert (
        diff_knu(radius=1e-9, other_radius=1e-8).units ==
        u.dimensionless
    )
    assert (
        diff_knu(radius=1e-9, other_radius=1e-8).magnitude ==
        pytest.approx(4, rel=1e0)
    )

    assert (
        diff_knu(radius=1e-9, other_radius=1e-8).magnitude <=
        diff_knu(radius=1e-9, other_radius=1e-9).magnitude
    )

    assert (
        diff_knu(
            radius=1e-9, other_radius=1e-8, charge=-1, other_charge=1
        ).magnitude >=
        diff_knu(
            radius=1e-9, other_radius=1e-8, charge=-1, other_charge=0
        ).magnitude
    )



"""test for import_interface.py"""


import tempfile
import os
import random
import string
import shutil
import numpy as np
from particula.data import loader_interface


def generate_files(
        file_count=10,
):
    """generate temp files for testing"""
    # create temp folder
    temp_folder = tempfile.mkdtemp(prefix='interfaces_')
    # create a subfolder
    subfolder_name = 'subfolder'
    subfolder_path = os.path.join(temp_folder, subfolder_name)
    os.mkdir(subfolder_path)

    size = 1000
    # create temp files
    file_names = []
    file_size = []
    for _ in range(file_count):
        file_name = ''.join(random.choices(string.ascii_uppercase +
                                           string.digits, k=10))
        file_names.append(file_name)
        file_size.append(size)
        # subfolder1
        file_path = os.path.join(subfolder_path, file_name)
        with open(file_path, 'wb') as f:
            f.write(np.zeros(size, dtype=np.uint8))

    return subfolder_name, temp_folder, file_size, file_names


def delete_temp_files(temp_folder):
    """delete temp files"""
    shutil.rmtree(temp_folder)


def test_get_new_files():
    """test for get_new_files function"""

    # generate temp files
    subfolder_name, temp_folder, _, file_names = generate_files()

    settings = {
        'relative_data_folder': subfolder_name,
        'filename_regex': '*',
        'MIN_SIZE_BYTES': 10,
    }
    # load files location
    full_paths, first_pass, file_info = loader_interface.get_new_files(
        path=temp_folder,
        import_settings=settings,
    )
    # check if the files are loaded
    assert first_pass is True
    assert len(full_paths) == len(file_names)
    assert len(file_info) == len(file_names)

    # remove a file from the info list and re-run, to check if the file loads
    last_file = file_info[-1]
    file_info.pop()
    full_paths, first_pass, file_info = loader_interface.get_new_files(
        path=temp_folder,
        import_settings=settings,
        loaded_list=file_info,
    )

    assert first_pass is False
    assert file_info[0] == last_file

    # delete temp files
    delete_temp_files(temp_folder)

    # single file test
    # create temp folder
    subfolder_name, temp_folder, _, file_names = generate_files(
        file_count=1)

    full_paths, first_pass, file_info = loader_interface.get_new_files(
        path=temp_folder,
        import_settings=settings,
    )

    assert first_pass is True
    assert len(full_paths) == len(file_names)
    assert len(file_info) == len(file_names)

    # retest with no new files

    full_paths, first_pass, file_info = loader_interface.get_new_files(
        path=temp_folder,
        import_settings=settings,
        loaded_list=file_info,
    )

    assert first_pass is False
    assert len(full_paths) == 0
    assert len(file_info) == 0



""" statics -> dynamics
"""

import numpy as np

from hypersolver.derivative import ord1_acc4

from particula.particle import Particle
from particula.util.coagulation_rate import CoagulationRate


class Rates:
    """ The class to calculate the rates
    """

    def __init__(
        self,
        particle=None,
        lazy=True,
    ):
        """ setting up the class
        """

        if particle is None or not isinstance(particle, Particle):
            raise ValueError("You must provide a baseline Particle object!")

        self.particle = particle
        self.lazy = lazy
        self.particle_distribution = self.particle.particle_distribution()
        self.particle_radius = self.particle.particle_radius
        self.particle_coagulation = self.particle.coagulation()
        self.particle_formation_rate = self.particle.particle_formation_rate

        if not self.lazy:
            self.eager_coags = CoagulationRate(
                distribution=self.particle_distribution,
                radius=self.particle_radius,
                kernel=self.particle_coagulation,
                lazy=self.lazy
            ).eager_coags

    def _coag_loss_gain(self):
        """ get both loss and gain
        """
        return CoagulationRate(
            distribution=self.particle_distribution,
            radius=self.particle_radius,
            kernel=self.particle_coagulation,
        )

    def coagulation_loss(self):
        """ get the coagulation loss rate
        """
        return self._coag_loss_gain().coag_loss() if self.lazy \
            else self.eager_coags[0]

    def coagulation_gain(self):
        """ get coagulation gain rate
        """
        return self._coag_loss_gain().coag_gain() if self.lazy \
            else self.eager_coags[1]

    def coagulation_rate(self):
        """ get the coagulation rate by summing the loss and gain rates
        """

        return self.coagulation_gain() - self.coagulation_loss()

    def condensation_growth_speed(self):
        """ condensation speed
        """

        return self.particle.particle_growth()

    def condensation_growth_rate(self):
        """ condensation rate
        """
        return ord1_acc4(
            -self.condensation_growth_speed().m * self.particle_distribution.m,
            self.particle_radius.m
        ) * (
            self.condensation_growth_speed().u * self.particle_distribution.u /
            self.particle_radius.u
        )

    def nucleation_rate(self):
        """ nucleation rate
        """
        result = np.zeros(
            self.condensation_growth_rate().m.shape
        )*self.particle_formation_rate.u
        result[0] = self.particle_formation_rate
        return result

    def dilution_rate(self):
        """ dilution rate
        """
        return (
            - self.particle.dilution_rate_coefficient() *
            self.particle_distribution
        )

    def wall_loss_rate(self):
        """ wall loss rate
        """
        return (
            - self.particle.wall_loss_coefficient() *
            self.particle_distribution
        )

    def sum_rates(
            self,
            coagulation=True,
            condensation=True,
            nucleation=True,
            dilution=False,
            wall_loss=False,
    ):  # pylint: disable=too-many-arguments
        """Sum rates, with options to disable individual rate terms.

        Args:
        ----------
        coagulation : bool, optional
            does the coagulation calcuation, by default True
        condensation : bool, optional
            does the condensation calculation, by default True
        nucleation : bool, optional
            does the nucleation calculation, by default True
        dilution : bool, optional
            does the dilution calculation, by default False
        wall_loss : bool, optional
            does the wall loss calculation, by default False

        """

        # Define a dictionary that maps option names to rate functions
        rates = {
            'coagulation': self.coagulation_rate,
            'condensation': self.condensation_growth_rate,
            'nucleation': self.nucleation_rate,
            'dilution': self.dilution_rate,
            'wall_loss': self.wall_loss_rate,
        }

        # Define a dictionary that maps option names to their Boolean values
        options = {
            'coagulation': coagulation,
            'condensation': condensation,
            'nucleation': nucleation,
            'dilution': dilution,
            'wall_loss': wall_loss,
        }
        # Return the sum of the rates that are enabled
        return sum(
            rates[key]() if option else 0 for key, option in options.items()
        )



""" calculate the condensation driving force

    https://www.nature.com/articles/nature18271
"""

from particula.util.kelvin_correction import kelvin_term as ktfun


def condensation_force(
    vapor_concentraton,
    sat_vapor_concentration,
    particle_activity=None,
    **kwargs
):
    """ calculate the condensation driving force

        Equation (9) in https://www.nature.com/articles/nature18271
    """
    fpa = particle_activity_fun(**kwargs) if particle_activity is None \
        else particle_activity

    return (
        vapor_concentraton -
        sat_vapor_concentration *
        fpa
    )


def particle_activity_fun(
    mass_fraction,
    activity_coefficient,
    kelvin_term=None,
    **kwargs
):
    """ calculate the particle activity

        Equation (9--10) in https://www.nature.com/articles/nature18271
    """

    ktf = ktfun(**kwargs) if kelvin_term is None else kelvin_term

    return (
        mass_fraction *
        activity_coefficient *
        ktf
    )



"""Contains tests for the collisions module."""
import torch
from particula.lagrangian import collisions


def test_find_collisions():
    """Test finding collisions between particles."""
    distance_matrix = torch.tensor([[-1, 2.0, 3.0],
                                    [1.0, -1, 3.0],
                                    [1.0, -0.1, -1]])
    indices = torch.tensor([0, 1, 2])
    mass = torch.tensor([1.0, 2.0, 3.0])
    k = 1
    expected_result = torch.tensor([[1, 2]], dtype=torch.int32)
    result = collisions.find_collisions(distance_matrix, indices, mass, k)
    assert torch.allclose(result, expected_result)


def test_coalescence():
    """Test coalescence of particles."""
    velocity = torch.tensor([[1.0, 2.0, 3.0],
                             [1.0, 2.0, 6.0],
                             [7.0, 8.0, 9.0],
                             [0.0, 0.0, 0.0]]).T
    mass = torch.tensor([1.0, 2.0, 3.0, 1.0])
    radius = torch.tensor([1.0, 2.0, 0.5, 0.5])
    position = torch.tensor([[1.0, 1.0, 2.0],
                             [1.0, 1.0, 3.0],
                             [7.0, 8.0, 9.0],
                             [0.0, 0.0, 0.0]]).T
    collision_indices_pairs = torch.tensor([[0, 1]], dtype=torch.int32)
    expected_velocity = torch.tensor([[1.0, 2.0, 5.0],
                                      [0.0, 0.0, 0.0],
                                      [7.0, 8.0, 9.0],
                                      [0.0, 0.0, 0.0]]).T
    expected_mass = torch.tensor([3.0, 0.0, 3.0, 1.0])
    result_velocity, result_mass = collisions.coalescence(
        position=position,
        velocity=velocity,
        mass=mass,
        radius=radius,
        collision_indices_pairs=collision_indices_pairs)
    assert torch.allclose(result_mass, expected_mass, atol=1e-4)
    assert torch.allclose(result_velocity, expected_velocity, atol=1e-4)




""" test the coag approx utility
"""

import pytest

from particula.util.approx_coagulation import approx_coag_less
from particula import u


def test_approx_coag_less():
    """ testing
    """

    assert approx_coag_less(
        diff_knu=1e-9, cpr=1, approx="dy2007").u == u.dimensionless
    assert approx_coag_less(
        diff_knu=1e-32, cpr=1, approx="dy2007").m == pytest.approx(0)

    assert approx_coag_less(
        diff_knu=1e-9, approx="hardsphere").u == u.dimensionless
    assert approx_coag_less(
        diff_knu=0, approx="hardsphere").m == pytest.approx(0)
    assert approx_coag_less(
        diff_knu=1e-32, approx="hardsphere").m == pytest.approx(0)

    assert approx_coag_less(
        diff_knu=1e-9, cpr=1, approx="gk2008").u == u.dimensionless
    assert approx_coag_less(
        diff_knu=1e-32, cpr=1, approx="gk2008").m == pytest.approx(0)

    assert approx_coag_less(
        diff_knu=1e-9, cpr=1, approx="gh2012").u == u.dimensionless
    assert approx_coag_less(
        diff_knu=0, cpr=1, approx="gh2012").m == pytest.approx(0)
    assert approx_coag_less(
        diff_knu=1e-32, cpr=1, approx="gh2012").m == pytest.approx(0)

    assert approx_coag_less(
        diff_knu=1e-9, cpr=1, approx="cg2019").u == u.dimensionless
    assert approx_coag_less(
        diff_knu=1e-32, cpr=1, approx="cg2019").m == pytest.approx(0)



""" testing the mean free path calculation
"""

import pytest
from particula import u
from particula.util.mean_free_path import mfp


def test_mfp():
    """ Testing the mean free path:

        1. test unitless and unit inputs
        2. test correct units
        3. test the calculated value (ref: ~66e-9 m mfp at sdt)
        4. test errors for invalid inputs

    """

    a_mfp = mfp(temperature=298*u.K, pressure=101325 * u.Pa)
    b_mfp = mfp(temperature=298, pressure=101325)
    c_mfp = mfp(temperature=298, pressure=101325, molecular_weight=0.03)

    assert a_mfp == b_mfp
    assert a_mfp.units == u.m
    assert a_mfp.magnitude == pytest.approx(66.4e-9, rel=1e-1)
    assert c_mfp <= a_mfp

    assert mfp(temperature=[200, 300]).m.shape == (2,)
    assert mfp(pressure=[1e5, 1.1e5]).m.shape == (2,)
    assert mfp(temperature=[200, 300], pressure=[1e5, 1.1e5]).m.shape == (2,)

    with pytest.raises(ValueError):
        mfp(temperature=5*u.m, pressure=101325*u.Pa)

    with pytest.raises(ValueError):
        mfp(temperature=298*u.K, pressure=5*u.m)

    with pytest.raises(ValueError):
        mfp(temperature=300*u.K,
            pressure=101325*u.Pa,
            molecular_weight=0.03*u.m/u.mol,
            )



""" calculating the surface tension of a mixture of solutes and water """
# pylint: disable=too-many-arguments

import numpy as np
from particula.util.input_handling import in_surface_tension, in_temperature, \
    in_scalar, in_volume, in_radius


def water(
        temperature,
        critical_temperature=647.15):
    """Calculate the surface tension of water using the equation from Kalova
    and Mares (2018).

    Args:
    ----------
        Temperature : float, Ambient temperature of air
        CritTemp : float, optional: Critical temperature of water

    Returns:
    -------
        sigma : float, Surface tension of water at the given temperature
    """
    temperature = in_temperature(temperature)
    critical_temperature = in_temperature(critical_temperature)
    # Dimensionless parameter from fitting equation
    tau = 1 - temperature / critical_temperature

    # Surface tension in mN/m
    sigma = 241.322 * (tau.m ** 1.26) * (
        1 - 0.0589 * (tau.m ** 0.5) - 0.56917 * tau.m
    )

    return in_surface_tension(sigma / 1000)


def dry_mixing(volume_fractions, surface_tensions):
    """Function to calculate the effective surface tension of a dry mixture.

    Args:
    ----------
    volume_fractions : array, volume fractions of solutes
    surface_tensions : array, surface tensions of solutes

    Returns:
    --------
    sigma : array, surface tension of droplet
    """
    volume_fractions = in_scalar(volume_fractions)
    surface_tensions = in_surface_tension(surface_tensions)
    if np.sum(volume_fractions) != 1:
        volume_fractions = volume_fractions / np.sum(volume_fractions)

    # Calculate the surface tension of the mixture
    return np.sum(volume_fractions * surface_tensions)


def wet_mixing(
    volume_solute,
    volume_water,
    wet_radius,
    surface_tension_solute,
    temperature,
    method='film'
):
    """Function to calculate the effective surface tension of a wet mixture.

    Args:
    ----------
    volume_solute : array, volume of solute mixture
    volume_water : array, volume of water
    surface_tension_solute : array, surface tension of solute mixture
    temperature : float, temperature of droplet
    method : str, optional: [film, volume] method to calculate effective
        surface tension

    Returns:
    --------
    EffSigma : array, effective surface tension of droplet
    """

    volume_solute = in_volume(volume_solute)
    volume_water = in_volume(volume_water)
    wet_radius = in_radius(wet_radius)
    surface_tension_solute = in_surface_tension(surface_tension_solute)
    temperature = in_temperature(temperature)

    # Organic-film method
    if method == 'film':

        # Characteristic monolayer thickness taken from AIOMFAC
        mono = in_radius(0.3e-9)

        # Volume of the monolayer
        film_volume = (4 / 3) * np.pi * (wet_radius **
                                         3 - (wet_radius - mono)**3)

        # Determine if there's enough organics to cover the particle
        surface_coverage = volume_solute / film_volume

        if surface_coverage >= 1:
            # If the particle is completely covered, the surface tension is
            # just the surface tension of the solute
            sigma = surface_tension_solute
        else:
            # If the particle is not completely covered, the surface tension
            # is a weighted average of the surface tension of the solute and
            # water
            sigma = surface_tension_solute * surface_coverage + \
                water(temperature) * (1 - surface_coverage)
    elif method == 'volume':
        # Volume method
        sigma = (
            volume_solute * surface_tension_solute +
            volume_water * water(temperature)
        ) / (volume_solute + volume_water)

    return sigma



"""creates the lake class, a collection of streams"""


from dataclasses import dataclass, field
from typing import Dict, Iterator, Any, Tuple
from particula.data.stream import Stream


@dataclass
class Lake:
    """A class representing a lake which is a collection of streams.

    Attributes:
        streams (Dict[str, Stream]): A dictionary to hold streams with their
        names as keys.
    """
    streams: Dict[str, Stream] = field(default_factory=dict)

    def add_stream(self, stream: Stream, name: str) -> None:
        """Add a stream to the lake.

        Args:
        -----------
            stream (Stream): The stream object to be added.
            name (str): The name of the stream.

        Raises:
        -------
            ValueError: If the stream name is already in use or not a valid
            identifier.
        """
        if name in self.streams:
            raise ValueError(f"Stream name {name} already in use")
        if not name.isidentifier():
            raise ValueError(
                f"Stream name '{name}' is not a valid python identifier")
        self.streams[name] = stream

    def __getattr__(self, name: str) -> Any:
        """Allow accessing streams as an attributes.
        Raises:
            AttributeError: If the stream name is not in the lake.
        Example: lake.stream_name
        """
        try:
            return self.streams[name]
        except KeyError as exc:
            raise AttributeError(f"No stream named {name}") from exc

    def __dir__(self) -> list:
        """List available streams.
        Example: dir(lake)"""
        return list(self.streams.keys())

    def __iter__(self) -> Iterator[Any]:
        """Iterate over the streams in the lake.
        Example: [stream.header for stream in lake]""
        """
        return iter(self.streams.items())

    def items(self) -> Iterator[Tuple[Any, Any]]:
        """Return an iterator over the key-value pairs."""
        return iter(self)

    def values(self) -> Iterator[Any]:
        """Return an iterator over the values."""
        return iter(self.streams.values())

    def keys(self) -> Iterator[Any]:
        """Return an iterator over the keys."""
        return iter(self.streams.keys())

    def __len__(self) -> int:
        """Return the number of streams in the lake.
        Example: len(lake)"""
        return len(self.streams)

    def __getitem__(self, key: str) -> Any:
        """Get a stream by name.
        Example: lake['stream_name']"""
        return self.streams[key]

    def __setitem__(self, key: str, value: Stream) -> None:
        """Set a stream by name.
        Example: lake['stream_name'] = new_stream"""
        # verify it is a stream object
        if not isinstance(value, Stream):
            raise ValueError(f"This is not a Stream object, {value}")
        self.streams[key] = value

    def __delitem__(self, key: str) -> None:
        """Remove a stream by name.
        Example: del lake['stream_name']"""
        if key in self.streams:
            del self.streams[key]
        else:
            raise ValueError(f"No stream named {key} to remove")

    def __repr__(self) -> str:
        """Return a string representation of the lake.
        Example: print(lake)"""
        return f"Lake with streams: {list(self.streams.keys())}"

    @property
    def summary(self) -> None:
        """Return a string summary iterating over each stream
            and print Stream.header.
        Example: lake.summary
        """
        for stream in self.streams:
            print(f"{stream} Headers:")
            print(f"      {self[stream].header}")



"""test the initialization of the Lake class."""


import numpy as np
from particula.data.lake import Lake
from particula.data.stream import Stream


def test_lake_initialization():
    """Test the initialization of the Lake class."""

    header = ['header11', 'header12']
    data = np.array([1, 2, 3])
    time = np.array([1.0, 2.0, 3.0])
    files = ['file1', 'file2']
    stream1 = Stream(header=header, data=data, time=time, files=files)

    header = ['header21', 'header22']
    data = np.array([1, 2, 3])
    time = np.array([1.0, 2.0, 3.0])
    files = ['file1', 'file2']
    stream2 = Stream(header=header, data=data, time=time, files=files)

    lake = Lake()
    lake.add_stream(stream1, name='BlueRiver')
    lake.add_stream(stream2, name='GreenRiver')

    assert lake.BlueRiver.header == stream1.header
    assert lake.GreenRiver.header == stream2.header



""" testing the DynamicStep class
"""

from particula import u
from particula.particle import Particle
from particula.rates import Rates


def test_dyns():
    """ test the dynamic step functionalities
    """

    some_particle = Particle(
        mode=100e-9,
        nbins=100,
        nparticles=1e5,
        gsigma=1.25,
        particle_formation_rate=0.5e8,
    )

    dyns = Rates(some_particle)

    assert dyns.coagulation_loss().u == u.m**-4/u.s
    assert dyns.coagulation_gain().u == u.m**-4/u.s
    assert dyns.coagulation_rate().u == u.m**-4/u.s
    assert dyns.condensation_growth_speed().u == u.m/u.s
    assert dyns.condensation_growth_rate().u == u.m**-4/u.s
    assert dyns.nucleation_rate().u == u.m**-4/u.s

    assert dyns.coagulation_loss().m.shape == (100,)
    assert dyns.coagulation_gain().m.shape == (100,)
    assert dyns.coagulation_rate().m.shape == (100,)
    assert dyns.condensation_growth_speed().m.shape == (100,)
    assert dyns.condensation_growth_rate().m.shape == (100,)
    assert dyns.nucleation_rate().m.shape == (100,)

    assert dyns.nucleation_rate().m[0] == 0.5e8
    assert dyns.nucleation_rate().m[1:].max() == 0



""" testing the input_handling function
"""

import pytest
from particula import u
from particula.util.input_handling import (in_density, in_length,
                                           in_molecular_weight, in_pressure,
                                           in_radius, in_scalar,
                                           in_temperature, in_viscosity,
                                           in_volume)


def test_in_temp():
    """ Testing the in_temperature function
    """

    temp = in_temperature(5)
    assert temp.units == u.K
    assert temp.magnitude == 5

    temp = in_temperature(5*u.K)
    assert temp.units == u.K
    assert temp.magnitude == 5

    temp = in_temperature(u.Quantity(5, u.degC))
    assert temp.units == u.K
    assert temp.magnitude == u.Quantity(5, u.degC).m_as("degK")

    temp = in_temperature(u.Quantity(5, u.degF))
    assert temp.units == u.K
    assert temp.magnitude == u.Quantity(5, u.degF).m_as("kelvin")

    with pytest.raises(ValueError):
        temp = in_temperature(5*u.m)


def test_in_vis():
    """ Testing the in_viscosity function
    """

    vis = in_viscosity(5)
    assert vis.units == u.kg / u.m / u.s
    assert vis.magnitude == 5

    vis = in_viscosity(5 * u.kg / u.m / u.s**2 * u.s)
    assert vis.units == u.kg / u.m / u.s
    assert vis.magnitude == 5

    vis = in_viscosity(u.Quantity(5, u.mPa * u.s))
    assert vis.units == u.kg / u.m / u.s
    assert vis.magnitude == u.Quantity(5, u.mPa * u.s).m_as("Pa*s")

    vis = in_viscosity(u.Quantity(5, u.kg / u.m / u.s))
    assert vis.units == u.kg / u.m / u.s
    assert vis.magnitude == u.Quantity(5, u.kg / u.m / u.s).m_as("Pa*s")

    vis = in_viscosity(u.Quantity(5, u.kg / u.m / u.s))
    assert vis.units == u.kg / u.m / u.s
    assert vis.magnitude == u.Quantity(5, u.kg / u.m / u.s).m_as("Pa*s")

    vis = in_viscosity(u.Quantity(5, u.mPa * u.s))
    assert vis.units == u.kg / u.m / u.s
    assert vis.magnitude == u.Quantity(5, u.mPa * u.s).m_as("Pa*s")

    vis = in_viscosity(u.Quantity(5, u.mPa * u.s))
    assert vis.units == u.kg / u.m / u.s
    assert vis.magnitude == u.Quantity(5, u.mPa * u.s).m_as("Pa*s")

    with pytest.raises(ValueError):
        vis = in_viscosity(5*u.m)


def test_in_molec_wt():
    """ Testing the in_molecular_weight function
    """

    molec_wt = in_molecular_weight(5)
    assert molec_wt.units == u.kg / u.mol
    assert molec_wt.magnitude == 5

    molec_wt = in_molecular_weight(5 * u.kg / u.mol)
    assert molec_wt.units == u.kg / u.mol
    assert molec_wt.magnitude == 5

    molec_wt = in_molecular_weight(u.Quantity(5, u.g / u.mol))
    assert molec_wt.units == u.kg / u.mol
    assert molec_wt.magnitude == u.Quantity(5, u.g / u.mol).m_as("kg/mol")

    molec_wt = in_molecular_weight(u.Quantity(5, u.kg / u.mol))
    assert molec_wt.units == u.kg / u.mol
    assert molec_wt.magnitude == u.Quantity(5, u.kg / u.mol).m_as("kg/mol")

    molec_wt = in_molecular_weight(u.Quantity(5, u.g / u.mol))
    assert molec_wt.units == u.kg / u.mol
    assert molec_wt.magnitude == u.Quantity(5, u.g / u.mol).m_as("kg/mol")

    with pytest.raises(ValueError):
        molec_wt = in_molecular_weight(5*u.m)


def test_in_pres():
    """ Testing the in_pressure function
    """

    pres = in_pressure(5)
    assert pres.units == u.kg / u.m / u.s**2
    assert pres.magnitude == 5

    pres = in_pressure(5 * u.kg / u.m / u.s**2)
    assert pres.units == u.kg / u.m / u.s**2
    assert pres.magnitude == 5

    pres = in_pressure(u.Quantity(5, u.mPa))
    assert pres.units == u.kg / u.m / u.s**2
    assert pres.magnitude == u.Quantity(5, u.mPa).m_as("Pa")

    pres = in_pressure(u.Quantity(5,  u.kg / u.m / u.s**2))
    assert pres.units == u.kg / u.m / u.s**2
    assert pres.magnitude == u.Quantity(5,  u.kg / u.m / u.s**2).m_as("Pa")

    pres = in_pressure(u.Quantity(5, u.mmHg))
    assert pres.units == u.kg / u.m / u.s**2
    assert pres.magnitude == u.Quantity(5, u.mmHg).m_as("Pa")

    pres = in_pressure(u.Quantity(5, u.atm))
    assert pres.units == u.kg / u.m / u.s**2
    assert pres.magnitude == u.Quantity(5, u.atm).m_as("Pa")

    pres = in_pressure(u.Quantity(5, u.torr))
    assert pres.units == u.kg / u.m / u.s**2
    assert pres.magnitude == u.Quantity(5, u.torr).m_as("Pa")

    pres = in_pressure(u.Quantity(5, u.bar))
    assert pres.units == u.kg / u.m / u.s**2
    assert pres.magnitude == u.Quantity(5, u.bar).m_as("Pa")

    pres = in_pressure(u.Quantity(5, u.mbar))
    assert pres.units == u.kg / u.m / u.s**2


def test_in_radius():
    """ Testing the in_radius function
    """

    radius = in_radius(5)
    assert radius.units == u.m
    assert radius.magnitude == 5

    radius = in_radius(5 * u.m)
    assert radius.units == u.m
    assert radius.magnitude == 5

    radius = in_radius(u.Quantity(5, u.mm))
    assert radius.units == u.m
    assert radius.magnitude == u.Quantity(5, u.mm).m_as("m")

    radius = in_radius(u.Quantity(5, u.cm))
    assert radius.units == u.m
    assert radius.magnitude == u.Quantity(5, u.cm).m_as("m")

    radius = in_radius(u.Quantity(5, u.m))
    assert radius.units == u.m
    assert radius.magnitude == u.Quantity(5, u.m).m_as("m")

    radius = in_radius(u.Quantity(5, u.km))
    assert radius.units == u.m
    assert radius.magnitude == u.Quantity(5, u.km).m_as("m")

    radius = in_radius(u.Quantity(5, u.inch))
    assert radius.units == u.m
    assert radius.magnitude == u.Quantity(5, u.inch).m_as("m")

    radius = in_radius(u.Quantity(5, u.ft))
    assert radius.units == u.m
    assert radius.magnitude == u.Quantity(5, u.ft).m_as("m")

    radius = in_radius(u.Quantity(5, u.mi))
    assert radius.units == u.m
    assert radius.magnitude == u.Quantity(5, u.mi).m_as("m")

    radius = in_radius(u.Quantity(5, u.nmi))
    assert radius.units == u.m
    assert radius.magnitude == u.Quantity(5, u.nmi).m_as("m")


def test_in_density():
    """ Testing the in_density function
    """

    density = in_density(5)
    assert density.units == u.kg / u.m**3
    assert density.magnitude == 5

    density = in_density(5 * u.kg / u.m**3)
    assert density.units == u.kg / u.m**3
    assert density.magnitude == 5

    density = in_density(u.Quantity(5, u.g / u.cm**3))
    assert density.units == u.kg / u.m**3
    assert density.magnitude == u.Quantity(5, u.g / u.cm**3).m_as("kg/m^3")

    density = in_density(u.Quantity(5, u.kg / u.m**3))
    assert density.units == u.kg / u.m**3
    assert density.magnitude == u.Quantity(5, u.kg / u.m**3).m_as("kg/m^3")

    density = in_density(u.Quantity(5, u.kg / u.cm**3))
    assert density.units == u.kg / u.m**3
    assert density.magnitude == u.Quantity(5, u.kg / u.cm**3).m_as("kg/m^3")

    density = in_density(u.Quantity(5, u.kg / u.cm**3))
    assert density.units == u.kg / u.m**3
    assert density.magnitude == u.Quantity(5, u.kg / u.cm**3).m_as("kg/m^3")

    density = in_density(u.Quantity(5, u.g / u.cm**3))
    assert density.units == u.kg / u.m**3
    assert density.magnitude == u.Quantity(5, u.g / u.cm**3).m_as("kg/m^3")

    density = in_density(u.Quantity(5, u.lb / u.ft**3))
    assert density.units == u.kg / u.m**3
    assert density.magnitude == u.Quantity(5, u.lb / u.ft**3).m_as("kg/m^3")


def test_in_scalar():
    """ Testing the in_scalar function
    """

    scalar = in_scalar(5)
    assert scalar.units == u.dimensionless
    assert scalar.magnitude == 5

    scalar = in_scalar(5 * u.dimensionless)
    assert scalar.units == u.dimensionless
    assert scalar.magnitude == 5

    with pytest.raises(ValueError):
        in_scalar(u.Quantity(5, u.m))


def test_in_length():
    """ Testing the in_length function
    """

    length = in_length(5)
    assert length.units == u.m
    assert length.magnitude == 5

    length = in_length(5 * u.m)
    assert length.units == u.m
    assert length.magnitude == 5

    length = in_length(u.Quantity(5, u.mm))
    assert length.units == u.m
    assert length.magnitude == u.Quantity(5, u.mm).m_as("m")

    length = in_length(u.Quantity(5, u.cm))
    assert length.units == u.m
    assert length.magnitude == u.Quantity(5, u.cm).m_as("m")

    length = in_length(u.Quantity(5, u.m))
    assert length.units == u.m
    assert length.magnitude == u.Quantity(5, u.m).m_as("m")

    length = in_length(u.Quantity(5, u.km))
    assert length.units == u.m
    assert length.magnitude == u.Quantity(5, u.km).m_as("m")

    length = in_length(u.Quantity(5, u.inch))
    assert length.units == u.m
    assert length.magnitude == u.Quantity(5, u.inch).m_as("m")

    length = in_length(u.Quantity(5, u.ft))
    assert length.units == u.m
    assert length.magnitude == u.Quantity(5, u.ft).m_as("m")

    length = in_length(u.Quantity(5, u.mi))
    assert length.units == u.m
    assert length.magnitude == u.Quantity(5, u.mi).m_as("m")

    length = in_length(u.Quantity(5, u.nmi))
    assert length.units == u.m
    assert length.magnitude == u.Quantity(5, u.nmi).m_as("m")


def test_in_volume():
    """ Testing the in_volume function
    """

    volume = in_volume(5)
    assert volume.units == u.m**3
    assert volume.magnitude == 5

    volume = in_volume(5 * u.m**3)
    assert volume.units == u.m**3
    assert volume.magnitude == 5

    volume = in_volume(u.Quantity(5, u.mm**3))
    assert volume.units == u.m**3
    assert volume.magnitude == u.Quantity(5, u.mm**3).m_as("m^3")

    volume = in_volume(u.Quantity(5, u.cm**3))
    assert volume.units == u.m**3
    assert volume.magnitude == u.Quantity(5, u.cm**3).m_as("m^3")

    volume = in_volume(u.Quantity(5, u.m**3))
    assert volume.units == u.m**3
    assert volume.magnitude == u.Quantity(5, u.m**3).m_as("m^3")

    volume = in_volume(u.Quantity(5, u.km**3))
    assert volume.units == u.m**3
    assert volume.magnitude == u.Quantity(5, u.km**3).m_as("m^3")

    volume = in_volume(u.Quantity(5, u.inch**3))
    assert volume.units == u.m**3
    assert volume.magnitude == u.Quantity(5, u.inch**3).m_as("m^3")

    volume = in_volume(u.Quantity(5, u.ft**3))
    assert volume.units == u.m**3
    assert volume.magnitude == u.Quantity(5, u.ft**3).m_as("m^3")

    volume = in_volume(u.Quantity(5, u.mi**3))
    assert volume.units == u.m**3
    assert volume.magnitude == u.Quantity(5, u.mi**3).m_as("m^3")



""" test knudsen number utility

    the knudsen number goes like:
        0   for larger  particles
        inf for smaller particles
"""

import pytest
from particula import u
from particula.util.knudsen_number import knu


def test_kn():
    """ test the knudsen number utility
    """

    radius = 1e-9 * u.m
    mfp_air = 66 * u.nm

    assert (
        knu(radius=radius, mfp=mfp_air) ==
        pytest.approx(66)
    )

    assert (
        knu(radius=1e-10) ==
        pytest.approx(664, rel=1e-1)
    )

    assert (
        knu(radius=1e-3) ==
        pytest.approx(5e-5, rel=1e0)
    )

    assert (
        knu(radius=1e-20) ==
        pytest.approx(5e12, rel=1e0)
    )

    assert knu(radius=[1, 2, 3]).m.shape == (3,)
    assert knu(radius=1, mfp=[1, 2, 3]).m.shape == (3, 1)
    assert knu(radius=[1, 2, 3], mfp=[1, 2, 3]).m.shape == (3, 3)
    assert knu(radius=[1, 2, 3], temperature=[1, 2, 3]).m.shape == (3, 3)
    assert knu(radius=[1, 2, 3], pressure=[1, 2, 3]).m.shape == (3, 3)
    assert knu(radius=[1, 2, 3], molecular_weight=[1, 2, 3]).m.shape == (3, 3)

    assert knu(
        radius=[1, 2, 3], temperature=[1, 2, 3], pressure=[1, 2, 3]
    ).m.shape == (3, 3)



""" diffusion coefficient
"""

# difp = bolk.* temp.* mobp; % diffusion of particle

from particula.util.input_handling import in_temperature, in_handling
from particula.constants import BOLTZMANN_CONSTANT
from particula.util.aerodynamic_mobility import pam


def pdc(
    temperature=298.15,
    bolk=BOLTZMANN_CONSTANT,
    pam_val=None,
    **kwargs
):
    """ particle diffusion coefficient
    """
    temperature = in_temperature(temperature)
    bolk = in_handling(bolk, BOLTZMANN_CONSTANT.u)
    pam_val = pam_val if pam_val is not None else pam(**kwargs)

    return (
        bolk * temperature * pam_val
    )



""" testing getting constants from the get_constants.py file
"""

from particula import u
from particula.constants import (AVOGADRO_NUMBER, BOLTZMANN_CONSTANT,
                                 GAS_CONSTANT)


def test_constants():
    """ simple tests are conducted as follows:

            * see if GAS_CONSTANT maniuplation is good
            * see if BOLTZMANN_CONSTANT units are good
    """

    assert (
        GAS_CONSTANT
        ==
        BOLTZMANN_CONSTANT * AVOGADRO_NUMBER
    )

    assert (
        GAS_CONSTANT.units
        ==
        BOLTZMANN_CONSTANT.units * AVOGADRO_NUMBER.units
    )

    assert (
        GAS_CONSTANT.magnitude
        ==
        BOLTZMANN_CONSTANT.magnitude * AVOGADRO_NUMBER.magnitude
    )

    assert (
        BOLTZMANN_CONSTANT.units
        ==
        u.m**2 * u.kg / u.s**2 / u.K
    )



"""Binary activity coefficient model for organic-water mixtures.

Gorkowski, K., Preston, T. C., &#38; Zuend, A. (2019).
Relative-humidity-dependent organic aerosol thermodynamics
Via an efficient reduced-complexity model.
Atmospheric Chemistry and Physics
https://doi.org/10.5194/acp-19-13383-2019
"""

from typing import Optional, Union, Tuple
from numpy.typing import ArrayLike
import numpy as np

from particula.activity.machine_limit import safe_exp
from particula.activity import phase_separation
from particula.activity.species_density import organic_density_estimate


# the fit values for the activity model
FIT_LOW = {'a1': [7.089476E+00, -7.711860E+00, -3.885941E+01, -1.000000E+02],
           'a2': [-6.226781E-01, -1.000000E+02, 3.081244E-09, 6.188812E+01],
           's': [-5.988895E+00, 6.940689E+00]}
FIT_MID = {'a1': [5.872214E+00, -4.535007E+00, -5.129327E+00, -2.809232E+01],
           'a2': [-9.740486E-01, -1.000000E+02, 2.109751E+00, -2.367683E+01],
           's': [-1.219164E+00, 4.742729E+00]}
FIT_HIGH = {'a1': [5.921550E+00, -2.528295E+00, -3.883017E+00, -7.898128E+00],
            'a2': [-1.000000E+02, -1.000000E+02, 1.353916E+00, -1.160145E+01],
            's': [-7.868187E-02, 3.650860E+00]}
# interpolation points, could be done smarter
INTERPOLATE_WATER_FIT = 500
LOWEST_ORGANIC_MOLE_FRACTION = 1e-12


def activity_coefficients(
    molar_mass_ratio: ArrayLike,
    organic_mole_fraction: ArrayLike,
    oxygen2carbon: ArrayLike,
    density: ArrayLike,
    functional_group=None,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray,
           np.ndarray, np.ndarray, np.ndarray]:
    """
    Calculate the activity coefficients for water and organic matter in
    organic-water mixtures.

    Args:
        - molar_mass_ratio: Ratio of the molecular weight of water to the
            molecular weight of organic matter.
        - organic_mole_fraction: Molar fraction of organic matter in the
            mixture.
        - oxygen2carbon: Oxygen to carbon ratio in the organic compound.
        - density: Density of the mixture.
        - functional_group: Optional functional group(s) of the organic
            compound, if applicable.

    Returns:
        A tuple containing the activity of water, activity
        of organic matter, mass fraction of water, and mass
        fraction of organic matter, gamma_water (activity coefficient),
        and gamma_organic (activity coefficient).
    """
    # check types
    organic_mole_fraction = np.asarray(organic_mole_fraction, dtype=np.float64)

    oxygen2carbon, molar_mass_ratio = convert_to_oh_equivalent(
        oxygen2carbon=oxygen2carbon,
        molar_mass_ratio=molar_mass_ratio,
        functional_group=functional_group
    )
    gibbs_mix, derivative_gibbs = gibbs_mix_weight(
        molar_mass_ratio=molar_mass_ratio,
        organic_mole_fraction=organic_mole_fraction,
        oxygen2carbon=oxygen2carbon,
        density=density,
    )
    # equations S8 S10
    # the func value for component 1 = LOG(activity coeff. water)
    ln_gamma_water = gibbs_mix - organic_mole_fraction * derivative_gibbs

    # the func value of the component 2 = LOG(activity coeff. of the organic)
    ln_gamma_org = gibbs_mix + (1.0 - organic_mole_fraction) * derivative_gibbs

    gamma_water = safe_exp(ln_gamma_water)
    gamma_organic = safe_exp(ln_gamma_org)

    activity_water = gamma_water * (1.0 - organic_mole_fraction)
    activity_organic = gamma_organic * organic_mole_fraction

    mass_water = (1.0 - organic_mole_fraction) * molar_mass_ratio / (
        (1.0 - organic_mole_fraction) * (molar_mass_ratio - 1) + 1
    )
    mass_organic = 1 - mass_water

    return activity_water, activity_organic, mass_water, mass_organic, \
        gamma_water, gamma_organic


def gibbs_of_mixing(
    molar_mass_ratio: ArrayLike,
    organic_mole_fraction: ArrayLike,
    oxygen2carbon: ArrayLike,
    density: ArrayLike,
    fit_dict: dict
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Calculate the Gibbs free energy of mixing for a binary mixture.

    Args:
        - molar_mass_ratio: The molar mass ratio of water to organic
            matter.
        - organic_mole_fraction: The fraction of organic matter.
        - oxygen2carbon: The oxygen to carbon ratio.
        - density: The density of the mixture.
        - fit_dict: A dictionary of fit values for the low oxygen2carbon region

    Returns:
        Tuple[np.ndarray, np.ndarray]: A tuple containing the Gibbs free
        energy of mixing and its derivative.
    """
    # check types
    molar_mass_ratio = np.asarray(molar_mass_ratio, dtype=np.float64)
    organic_mole_fraction = np.asarray(organic_mole_fraction, dtype=np.float64)
    oxygen2carbon = np.asarray(oxygen2carbon, dtype=np.float64)
    density = np.asarray(density, dtype=np.float64)

    c1 = coefficients_c(molar_mass_ratio, oxygen2carbon, fit_dict['a1'])
    c2 = coefficients_c(molar_mass_ratio, oxygen2carbon, fit_dict['a2'])

    rhor = 0.997 / density  # assumes water is the other fluid

    # equation S3
    # the scaled molar mass ratio of this mixture's components.
    scaled_molar_mass_ratio = molar_mass_ratio * fit_dict['s'][1] \
        * (1.0 + oxygen2carbon) ** fit_dict['s'][0]

    # phi2 is a scaled volume fraction
    phi2 = organic_mole_fraction / (
        organic_mole_fraction + (1.0 - organic_mole_fraction)
        * scaled_molar_mass_ratio / rhor
    )

    # equation S4
    sum1 = c1 + c2 * (1 - 2 * phi2)
    gibbs_mix = phi2 * (1.0 - phi2) * sum1

    # equation s6 the derivative of phi2 with respect to organic x2
    dphi2dx2 = (scaled_molar_mass_ratio / rhor) \
        * (phi2 / organic_mole_fraction)**2

    # equation S7
    derivative_gibbs_mix = (
        (1.0 - 2.0 * phi2) * sum1 - 2 * c2 * phi2 * (1.0 - phi2)
    ) * dphi2dx2

    return gibbs_mix, derivative_gibbs_mix


def gibbs_mix_weight(
    molar_mass_ratio: ArrayLike,
    organic_mole_fraction: ArrayLike,
    oxygen2carbon: ArrayLike,
    density: ArrayLike,
    functional_group: Optional[str] = None,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Gibbs free energy of mixing, see Gorkowski (2019), with weighted
    oxygen2carbon regions. Only can run one compound at a time.

    Args:
        - molar_mass_ratio: The molar mass ratio of water to organic
            matter.
        - organic_mole_fraction: The fraction of organic matter.
        - oxygen2carbon: The oxygen to carbon ratio.
        - density: The density of the mixture.
        - functional_group: Optional functional group(s) of the organic
            compound, if applicable.

    Returns:
        - gibbs_mix : Gibbs energy of mixing (including 1/RT)
        - derivative_gibbs : derivative of Gibbs energy with respect to
        - mole fraction of organics (includes 1/RT)
    """
    # check types
    density = np.asarray(density, dtype=np.float64)

    oxygen2carbon, molar_mass_ratio = convert_to_oh_equivalent(
        oxygen2carbon=oxygen2carbon,
        molar_mass_ratio=molar_mass_ratio,
        functional_group=functional_group
    )

    weights = bat_blending_weights(
        molar_mass_ratio=molar_mass_ratio,
        oxygen2carbon=oxygen2carbon
    )

    if weights[1] > 0:  # if mid region is used
        gibbs_mix_mid, derivative_gibbs_mid = gibbs_of_mixing(
            molar_mass_ratio=molar_mass_ratio,
            organic_mole_fraction=organic_mole_fraction,
            oxygen2carbon=oxygen2carbon,
            density=density,
            fit_dict=FIT_MID
        )

        if weights[0] > 0:  # if paired with low oxygen2carbon region
            gibbs_mix_low, derivative_gibbs_low = gibbs_of_mixing(
                molar_mass_ratio=molar_mass_ratio,
                organic_mole_fraction=organic_mole_fraction,
                oxygen2carbon=oxygen2carbon,
                density=density,
                fit_dict=FIT_LOW
            )
            gibbs_mix = weights[0] * gibbs_mix_low + weights[1] * gibbs_mix_mid
            derivative_gibbs = weights[0] * derivative_gibbs_low \
                + weights[1] * derivative_gibbs_mid
        else:  # else paired with high oxygen2carbon region
            gibbs_mix_high, derivative_gibbs_high = gibbs_of_mixing(
                molar_mass_ratio=molar_mass_ratio,
                organic_mole_fraction=organic_mole_fraction,
                oxygen2carbon=oxygen2carbon,
                density=density,
                fit_dict=FIT_HIGH
            )
            gibbs_mix = weights[2] * gibbs_mix_high + \
                weights[1] * gibbs_mix_mid
            derivative_gibbs = weights[2] * derivative_gibbs_high \
                + weights[1] * derivative_gibbs_mid
    else:  # when only high 2OC region is used
        gibbs_mix, derivative_gibbs = gibbs_of_mixing(
            molar_mass_ratio=molar_mass_ratio,
            organic_mole_fraction=organic_mole_fraction,
            oxygen2carbon=oxygen2carbon,
            density=density,
            fit_dict=FIT_HIGH
        )
    return gibbs_mix, derivative_gibbs


def biphasic_water_activity_point(
    oxygen2carbon: ArrayLike,
    hydrogen2carbon: ArrayLike,
    molar_mass_ratio: ArrayLike,
    functional_group: Optional[Union[list[str], str]] = None
) -> np.ndarray:
    """
    This function computes the biphasic to single phase
    water activity (RH*100).

    Args:
        - oxygen2carbon: The oxygen to carbon ratio.
        - hydrogen2carbon: The hydrogen to carbon ratio.
        - molar_mass_ratio: The molar mass ratio of water to organic
            matter.
        - functional_group: Optional functional group(s) of the organic
            compound, if applicable.

    Returns:
        - np.array: The RH cross point array.
    """
    # check types
    oxygen2carbon = np.asarray(oxygen2carbon, dtype=np.float64)
    hydrogen2carbon = np.asarray(hydrogen2carbon, dtype=np.float64)
    molar_mass_ratio = np.asarray(molar_mass_ratio, dtype=np.float64)
    # if 0-d then expand to 1-d
    if oxygen2carbon.ndim == 0:
        oxygen2carbon = np.expand_dims(oxygen2carbon, axis=0)
    if hydrogen2carbon.ndim == 0:
        hydrogen2carbon = np.expand_dims(hydrogen2carbon, axis=0)
    if molar_mass_ratio.ndim == 0:
        molar_mass_ratio = np.expand_dims(molar_mass_ratio, axis=0)

    water_activity_cross_point = np.zeros_like(oxygen2carbon)

    interpolate_step_numb = 200  # interpolation points
    # mole_frac = np.linspace(1e-12, 1, interpolate_step_numb + 1)
    mole_frac = np.logspace(-6, 0, interpolate_step_numb + 1)

    for i, _ in enumerate(oxygen2carbon):
        density = organic_density_estimate(
            molar_mass_ratio[i],
            oxygen2carbon[i],
            hydrogen2carbon[i],
            mass_ratio_convert=True)
        activities = activity_coefficients(
            molar_mass_ratio=molar_mass_ratio[i],
            organic_mole_fraction=mole_frac,
            oxygen2carbon=oxygen2carbon[i],
            density=density,
            functional_group=functional_group
        )

        if np.isnan(activities[0]).any():
            raise ValueError('water activity is NaN, check inputs')

        phase_check = phase_separation.find_phase_separation(
            activities[0],
            activities[1])

        if phase_check['phase_sep_check'] == 1:
            water_activity_cross_point[i] = phase_check['upper_seperation']
        else:
            water_activity_cross_point[i] = 0  # no phase separation

    # Checks outputs with in physical limits
    # round to zero
    water_activity_cross_point[water_activity_cross_point < 0] = 0
    # round max to 1
    water_activity_cross_point[water_activity_cross_point > 1] = 1

    return water_activity_cross_point


def convert_to_oh_equivalent(
    oxygen2carbon: ArrayLike,
    molar_mass_ratio: ArrayLike,
    functional_group: Optional[Union[list[str], str]] = None
) -> Tuple[np.ndarray, np.ndarray]:
    """
    just a pass through now, but will
    add the oh equivalent conversion
    """
    # check types
    oxygen2carbon = np.asarray(oxygen2carbon, dtype=np.float64)
    molar_mass_ratio = np.asarray(molar_mass_ratio, dtype=np.float64)

    # sourcery skip
    if functional_group is None:
        return oxygen2carbon, molar_mass_ratio
    if functional_group == 'alcohol':
        return oxygen2carbon + 1, molar_mass_ratio + 16  # fix this from SI
    raise ValueError('BAT functional group not recognized')


def bat_blending_weights(
        molar_mass_ratio: ArrayLike,
        oxygen2carbon: ArrayLike
) -> np.ndarray:
    """
    Function to estimate the blending weights for the BAT model.

    Args:
        - molar_mass_ratio: The molar mass ratio of water to organic
            matter.
        - oxygen2carbon: The oxygen to carbon ratio.

    Returns:
        - blending_weights : List of blending weights for the BAT model
        in the low, mid, and high oxygen2carbon regions.
    """
    # check types
    molar_mass_ratio = np.asarray(molar_mass_ratio, dtype=np.float64)
    oxygen2carbon = np.asarray(oxygen2carbon, dtype=np.float64)

    oxygen2carbon_ml = phase_separation.organic_water_single_phase(
        molar_mass_ratio=molar_mass_ratio)

    blending_weights = np.zeros(3)  # [low, mid, high] oxygen2carbon regions

    # lower to middle oxygen2carbon region
    if oxygen2carbon <= oxygen2carbon_ml * 0.75:
        b_ml = 0.189974476118418
        b_1 = 79.2606902175984
        b_2 = 0.0604293454322489

        oxygen2carbon_b = oxygen2carbon - oxygen2carbon_ml * b_ml
        weight_b = 1 / (1 + np.exp(
            - b_1 * (oxygen2carbon_b - b_2)
        ))  # logistic transfer function

        oxygen2carbon_b_norm = oxygen2carbon - (0.75 * oxygen2carbon_ml * b_ml)

        weight_norm = 1 / (1 + np.exp(
            - b_1 * (oxygen2carbon_b_norm - b_2)
        ))

        blending_weights[1] = weight_b / weight_norm
        blending_weights[0] = 1 - blending_weights[1]

    # middle to high oxygen2carbon region
    elif oxygen2carbon <= oxygen2carbon_ml * 2:
        b_1 = 75.0159268221068
        b_2 = 0.000947111285750515

        oxygen2carbon_b = oxygen2carbon - oxygen2carbon_ml
        blending_weights[2] = 1 / (1 + np.exp(
            - b_1 * (oxygen2carbon_b - b_2)
        ))  # logistic transfer function

        blending_weights[1] = 1 - blending_weights[2]

    else:  # high only region
        blending_weights[2] = 1

    return blending_weights


def coefficients_c(
    molar_mass_ratio: ArrayLike,
    oxygen2carbon: ArrayLike,
    fit_values: ArrayLike
) -> np.ndarray:
    """
    Coefficients for activity model, see Gorkowski (2019). equation S1 S2.

    Args:
        - molar_mass_ratio: The molar mass ratio of water to organic
            matter.
        - oxygen2carbon: The oxygen to carbon ratio.
        - fit_values: The fit values for the activity model.

    Returns:
        - np.ndarray: The coefficients for the activity model.
    """
    # check types
    molar_mass_ratio = np.asarray(molar_mass_ratio, dtype=np.float64)
    oxygen2carbon = np.asarray(oxygen2carbon, dtype=np.float64)
    fit_values = np.asarray(fit_values, dtype=np.float64)

    return fit_values[0] * np.exp(fit_values[1] * oxygen2carbon) + fit_values[
        2
    ] * np.exp(fit_values[3] * molar_mass_ratio)


def fixed_water_activity(
        water_activity: ArrayLike,
        molar_mass_ratio: ArrayLike,
        oxygen2carbon: ArrayLike,
        density: ArrayLike,
) -> Tuple:
    # pylint: disable=too-many-locals
    """
    Calculate the activity coefficients of water and organic matter in
    organic-water mixtures.

    This function assumes a fixed water activity value (e.g., RH = 75%
    corresponds to 0.75 water activity in equilibrium).
    It calculates the activity coefficients for different phases and
    determines phase separations if they occur.

    Parameters:
    water_activity (ArrayLike): An array of water activity values.
    molar_mass_ratio (ArrayLike): Array of molar mass ratios of the components.
    oxygen2carbon (ArrayLike): Array of oxygen-to-carbon ratios.
    density (ArrayLike): Array of densities of the mixture.

    Returns:
    Tuple: A tuple containing the activity coefficients for alpha and beta
            phases, and the alpha phase mole fraction.
           If no phase separation occurs, the beta phase values are None.
    """

    # check types
    water_activity = np.asarray(water_activity, dtype=np.float64)
    molar_mass_ratio = np.asarray(molar_mass_ratio, dtype=np.float64)
    oxygen2carbon = np.asarray(oxygen2carbon, dtype=np.float64)
    density = np.asarray(density, dtype=np.float64)

    # must have activity of water in increasing order
    if water_activity.size > 1 and water_activity[0] > water_activity[-1]:
        water_activity = np.flip(water_activity)
        flip = True
    else:
        flip = False

    organic_mole_fraction_array = np.linspace(
        1,
        LOWEST_ORGANIC_MOLE_FRACTION,
        INTERPOLATE_WATER_FIT,
        dtype=np.float64
    )

    # activity calculation
    activities = activity_coefficients(
        molar_mass_ratio=molar_mass_ratio,
        organic_mole_fraction=organic_mole_fraction_array,
        oxygen2carbon=oxygen2carbon,
        density=density
    )
    # find phase separation
    phase_check = phase_separation.find_phase_separation(
        activities[0],
        activities[1]
    )
    # ensure water activity type is float
    activities_water = np.asarray(activities[0], dtype=np.float64)
    if phase_check['phase_sep_check'] == 0:
        alpha_organic_mole_fraction = np.interp(
            xp=activities_water,
            fp=organic_mole_fraction_array,
            x=water_activity,
            left=1.0,
            right=LOWEST_ORGANIC_MOLE_FRACTION,
        )
        # activity calculation for alpha phase
        activities_alpha = activity_coefficients(
            molar_mass_ratio=molar_mass_ratio,
            organic_mole_fraction=alpha_organic_mole_fraction,
            oxygen2carbon=oxygen2carbon,
            density=density
        )
        activities_beta = None
        q_alpha = np.ones(water_activity.size)
        # change back to original order
        if flip:
            activities_alpha = np.flip(activities_alpha)
            q_alpha = np.flip(q_alpha)
        return (activities_alpha, activities_beta, q_alpha)

    # else phase separation occurs
    # split the activities into alpha and beta phases

    # alpha water rich phase
    alpha_water_activity = activities_water[
        phase_check['upper_seperation_index']:]
    alpha_organic_mole_fraction = organic_mole_fraction_array[
        phase_check['upper_seperation_index']:]
    # beta organic rich phase
    beta_water_activity = activities_water[
        : phase_check['matching_upper_seperation_index']
    ]
    beta_organic_mole_fraction = organic_mole_fraction_array[
        : phase_check['matching_upper_seperation_index']
    ]

    # find the water activity of the alpha phase
    alpha_organic_mole_fraction_interp = np.interp(
        xp=alpha_water_activity,
        fp=alpha_organic_mole_fraction,
        x=water_activity,
        left=np.nan,
        right=LOWEST_ORGANIC_MOLE_FRACTION,
    )
    # find the water activity of the beta phase
    beta_organic_mole_fraction_interp = np.interp(
        xp=beta_water_activity,
        fp=beta_organic_mole_fraction,
        x=water_activity,
        left=1,
        right=np.nan,
    )
    # calculate the activity coefficients for the alpha phase
    activities_alpha = activity_coefficients(
        molar_mass_ratio=molar_mass_ratio,
        organic_mole_fraction=alpha_organic_mole_fraction_interp,
        oxygen2carbon=oxygen2carbon,
        density=density
    )
    # calculate the activity coefficients for the beta phase
    activities_beta = activity_coefficients(
        molar_mass_ratio=molar_mass_ratio,
        organic_mole_fraction=beta_organic_mole_fraction_interp,
        oxygen2carbon=oxygen2carbon,
        density=density
    )
    q_alpha = phase_separation.q_alpha(
        seperation_activity=phase_check['upper_seperation'],
        activities=water_activity,
    )
    # change back to original order
    if flip:
        activities_alpha = np.flip(activities_alpha)
        activities_beta = np.flip(activities_beta)
        q_alpha = np.flip(q_alpha)

    return (activities_alpha, activities_beta, q_alpha)



""" statics -> dynamics
"""
from scipy.integrate import odeint, solve_ivp

from particula.rates import Rates


class Solver(Rates):
    """ dynamic solver
    """

    def __init__(
        self,
        time_span=None,
        do_coagulation=True,
        do_condensation=True,
        do_nucleation=True,
        do_dilution=False,
        do_wall_loss=False,
        **kwargs
    ):  # pylint: disable=too-many-arguments
        """ constructor
        """
        super().__init__(**kwargs)

        if time_span is None:
            raise ValueError("You must provide a time span!")

        self.time_span = time_span
        self.do_coagulation = do_coagulation
        self.do_condensation = do_condensation
        self.do_nucleation = do_nucleation
        self.do_dilution = do_dilution
        self.do_wall_loss = do_wall_loss

    def _ode_func(self, _nums, _,):
        """ ode_func
        """
        setattr(
            self,
            'particle_distribution',
            _nums*self.particle_distribution.u
        )

        return self.sum_rates(
            coagulation=self.do_coagulation,
            condensation=self.do_condensation,
            nucleation=self.do_nucleation,
            dilution=self.do_dilution,
            wall_loss=self.do_wall_loss,
        ).m

    def solution(
            self,
            method='odeint',
            **kwargs_ode
    ):
        """ solve the equation
        """
        if method == 'odeint':
            return odeint(
                func=self._ode_func,
                y0=self.particle.particle_distribution().m,
                t=self.time_span,
                **kwargs_ode
            )*self.particle.particle_distribution().u

        if method == 'solve_ivp':
            if 'method' not in kwargs_ode:
                kwargs_ode['method'] = 'BDF'  # Choose an appropriate method
            return solve_ivp(
                fun=self._ode_func,
                t_span=(self.time_span[0], self.time_span[-1]),
                y0=self.particle.particle_distribution().m,
                t_eval=self.time_span,
                **kwargs_ode
            ).y.T*self.particle.particle_distribution().u

        raise ValueError("Invalid method!")



"""Boundary conditions for Lagrangian particle simulations."""

import torch


def wrapped_cube(
        position: torch.Tensor, cube_side: float) -> torch.Tensor:
    """
    Apply cubic boundary conditions with wrap-around, to the position tensor.

    This function modifies positions that exceed the cubic domain side,
    wrapping them around to the opposite side of the domain. It handles both
    positive and negative overflows. The center of the cube is assumed to be
    at zero. If a particle is way outside the cube, it is wrapped around to
    the opposite side of the cube.

    Parameters:
    - position (torch.Tensor): A tensor representing positions that might
        exceed the domain boundaries. [3, num_particles]
    - cube_side (float): The cube side length of the domain.

    Returns:
    - torch.Tensor: The modified position tensor with boundary conditions
        applied.

    Example:
    position = torch.tensor([...])  # Position tensor
    cube_side = 10.0  # Define the domain
    wrapped_position = boundary.wrapped_cube(position,
        cube_side)
    """
    half_cube_side = cube_side / 2
    # Wrap around for positive overflow
    position = torch.where(
        position > half_cube_side,
        position - cube_side,
        position)

    # Wrap around for negative overflow
    position = torch.where(
        position < -half_cube_side,
        position + cube_side,
        position)

    # Wrap around for very large positive overflow
    position = torch.where(
        position > half_cube_side,
        -half_cube_side,
        position)

    # Wrap around for very large negative overflow
    position = torch.where(
        position < -half_cube_side,
        half_cube_side,
        position)

    return position



""" test the simple solver
"""

import pytest
import numpy as np
from particula import particle, dynamics

some_kwargs = {
    "vapor_concentration": 0.025e-9,
    "mode": 20e-8,  # 200 nm median
    "nbins": 1000,  # 1000 bins
    "cutoff": .9999999,
}

particle_dist = particle.Particle(**some_kwargs)

Solver = dynamics.Solver(
    particle=particle_dist,
    time_span=np.linspace(0, 100, 200),
)

FinerSolver = dynamics.Solver(
    particle=particle_dist,
    time_span=np.linspace(0, 100, 500),
)

radius = particle_dist.particle_radius.m
solution = Solver.solution()
fine_sols = FinerSolver.solution()


def test_solution():
    """ test fidelity of the solution
    """

    assert solution.m.shape == (200, 1000)
    assert solution.u == particle_dist.particle_distribution().u
    assert fine_sols.m.shape == (500, 1000)
    assert (solution.m[0, :] == fine_sols.m[0, :]).all()


def test_conservation():
    """ test conservation of mass/volume
    """

    assert (
        np.trapz((solution.m[0, :]-solution.m[-1, :])*radius**3, radius) /
        np.trapz((solution.m[0, :])*radius**3, radius)
        ==
        pytest.approx(0, abs=5e-3)
    )

    assert (
        np.trapz((fine_sols.m[0, :]-fine_sols.m[-1, :])*radius**3, radius) /
        np.trapz((fine_sols.m[0, :])*radius**3, radius)
        ==
        pytest.approx(0, abs=5e-3)  # allow for 0.5% each direction (1% total)
    )



"""stats module for datacula"""
# linting disabled until reformatting of this file


from typing import Tuple, Optional
import numpy as np


def merge_formatting(
        data_current: np.ndarray,
        header_current: list,
        data_new: np.ndarray,
        header_new: list
) -> Tuple[np.ndarray, list, np.ndarray, list]:
    """
    Formats two data arrays and their headers so that the data new can be
    subsiqently added to data current.

    Args:
        data_current (np.ndarray): First data array to merge.
        header_current (list): Header for the first data array.
        data_new (np.ndarray): Second data array to merge.
        header_new (list): Header for the second data array.

    Returns:
        (np.ndarray, list, np.array, list): A tuple formatted data
        and headers.

    Raises:
        ValueError: If the data arrays are not the same shape.
        ValueError: If the headers are not the same length.
    """
    # elements in header_new that are not in header_current
    header_new_not_listed = [
        x for x in header_new if x not in header_current
    ]
    header_list_not_new = [
        x for x in header_current if x not in header_new
    ]

    # expand the data array to include the new columns if there are any
    if bool(header_new_not_listed):
        header_current.extend(
            header_new_not_listed
        )
        # add other rows to the data array
        data_current = np.concatenate(
            (
                data_current,
                np.full((
                        data_current.shape[0],
                        len(header_new_not_listed)
                        ), np.nan)
            ),
            axis=1
        )

    if bool(header_list_not_new):
        header_new.extend(
            header_list_not_new
        )
        data_new = np.concatenate(
            (
                data_new,
                np.full((
                        data_new.shape[0],
                        len(header_list_not_new)
                        ), np.nan)
            ),
            axis=1
        )

    # check that the shapes are the same
    if data_current.shape[1] != data_new.shape[1]:
        raise ValueError(
            'data_current  ',
            data_current.shape,
            ' and data_new ',
            data_new.shape,
            ' are not the same shape, check the data formatting'
        )
    if len(header_current) != len(header_new):
        raise ValueError(
            'header_current ',
            len(header_current),
            ' and header_new ',
            len(header_new),
            ' are not the same shape, check the data formatting'
        )
    # check if the headers are numeric
    if np.all([x[0].isnumeric() for x in header_current]):
        # convert the headers to floats
        header_list_numeric = np.array(header_current).astype(float)
        header_new_numeric = np.array(header_new).astype(float)
        # find the indices of the headers
        header_stored_indices = np.argsort(header_list_numeric)
        header_new_indices = np.argsort(header_new_numeric)
        # sort the header and keep it a list
        header_current = [header_current[i] for i in header_stored_indices]
        header_new = [header_new[i] for i in header_new_indices]

        # sort the data
        data_current = data_current[:, header_stored_indices]
    else:
        # match header_new to header_current and sort the data
        header_new_indices = [
            header_new.index(x) for x in header_current
        ]
        header_new = [header_new[i] for i in header_new_indices]
    data_new = data_new[:, header_new_indices]
    return data_current, header_current, data_new, header_new


def average_to_interval(
    time_raw: np.ndarray,
    data_raw: np.ndarray,
    average_interval: float,
    average_interval_array: np.ndarray,
    average_data: np.ndarray,
    average_data_std: np.ndarray
) -> Tuple[np.ndarray, np.ndarray]:
    # pylint: disable=too-many-arguments, too-many-branches
    """
    Calculate the average of the data stream over the specified time intervals.

    This function calculates the average of the data stream over a series of
    time intervals specified by `average_interval_array`. The average and
    standard
    deviation of the data are calculated for each interval, and the results
    are returned as two arrays.

    Args:
    ----------
        time_raw (np.ndarray): An array of timestamps, sorted in ascending
            order.
        average_interval (float): The length of each time interval in seconds.
        average_interval_array (np.ndarray): An array of timestamps
            representing
            the start times of each time interval.
        data_raw (np.ndarray): An array of data points corresponding to the
            timestamps in `time_raw`.
        average_data (np.ndarray): An empty array of shape
            (num_channels, num_intervals)that will be filled with the
            average data for each time interval.
        average_data_std (np.ndarray): An empty array of shape
            (num_channels, num_intervals) that will be filled with the standard
            deviation of the data for each time interval.

    Returns:
    -------
        Tuple[np.ndarray, np.ndarray]: A tuple containing the average data
            and the standard deviation of the data, both as arrays of shape
            (num_channels, num_intervals).
    """
    # average the data in the time interval initialization
    # find closet index in time_raw to the start time
    start_index = np.argmin(np.abs(average_interval_array[0] - time_raw))
    # start_index = 0
    stop_index = 0
    interval_look_buffer_multiple = 2
    start_time = time_raw[0]

    # estimating how much of the time_raw we would need to look at for a
    # given average interval.
    if len(time_raw) > 100:
        time_lookup_span = round(
            (
                average_interval
                * interval_look_buffer_multiple
                / np.nanmean(np.diff(time_raw[:100]))
            )
        )
    else:
        time_lookup_span = 100

    # loop through the average time intervals
    for i, time_i in enumerate(average_interval_array, start=1):
        if (stop_index < len(time_raw)) and (start_time < time_i):

            # trying to only look at the time_raw in the average time
            # interval, assumes that the time stream is sorted
            if start_index + time_lookup_span < len(time_raw):
                compare_bool = np.nonzero(
                    time_raw[start_index:start_index + time_lookup_span]
                    >= time_i
                )
                if len(compare_bool[0]) > 0:
                    stop_index = start_index + compare_bool[0][0]
                else:
                    # used it all for this iteration
                    compare_bool = np.nonzero(
                        time_raw[start_index:]
                        >= time_i
                    )
                    stop_index = start_index + compare_bool[0][0]
                    # re-calculate time look up span,
                    # as timesteps have changed
                    if len(time_raw[start_index:]) > 100:
                        time_lookup_span = round(
                            average_interval
                            * interval_look_buffer_multiple
                            / np.nanmean(
                                np.diff(
                                    time_raw[
                                        start_index:start_index + 100
                                    ]
                                )
                            )
                        )
                    else:
                        time_lookup_span = 100
            else:
                compare_bool = np.nonzero(
                    time_raw[
                        start_index:start_index + time_lookup_span
                    ] >= time_i
                )
                if len(compare_bool[0]) > 0:
                    stop_index = start_index + compare_bool[0][0]
                else:
                    stop_index = len(time_raw)

            if start_index < stop_index:
                # average the data in the time interval
                average_data[i - 1, :] = np.nanmean(
                    data_raw[start_index:stop_index, :], axis=0
                )  # the actual averaging of data is here
                average_data_std[i - 1, :] = np.nanstd(
                    data_raw[start_index:stop_index, :], axis=0
                )  # the actual std data is here
            else:
                start_time = time_raw[stop_index]
            start_index = stop_index

    return average_data, average_data_std


def mask_outliers(
        data: np.ndarray,
        bottom: Optional[float] = None,
        top: Optional[float] = None,
        value: Optional[float] = None,
        invert: Optional[bool] = False
) -> np.ndarray:
    """
    Create a boolean mask for outliers in a data array. Outliers are defined as
    values that are either above or below a specified threshold, or that are
    equal to a specified value. Not all parameters need to be specified. If
    `invert` is True, the mask will be inverted. The mask will be True for
    False for outliers and True for non-outliers.

    Args:
    ----------
        data (np.ndarray): The data array to be masked.
        bottom (float): The lower threshold for outliers.
        top (float): The upper threshold for outliers.
        value (float): The value to be masked.
        invert (bool): If True, the mask will be inverted.

    Returns:
    -------
        np.ndarray: A boolean mask for the outliers in the data array. Mask is
            True for non-outliers and False for outliers, and the same shape as
            the data array.
    """

    # initialize the mask
    mask = np.zeros(data.shape, dtype=bool)

    # mask values below the bottom threshold
    if bottom is not None:
        mask = np.logical_or(mask, data < bottom)

    # mask values above the top threshold
    if top is not None:
        mask = np.logical_or(mask, data > top)

    # mask values equal to the specified value
    if value is not None:
        mask = np.logical_or(mask, data == value)

    # if you want true = outliers
    if invert:
        return mask

    # invert the mask to get true= non-outliers and false=outliers
    mask = np.logical_not(mask)
    return mask


def distribution_integration(
        distribution: np.ndarray,
        x_array: Optional[np.ndarray] = None,
        axis: int = 0
) -> np.ndarray:
    """
    Performs either PDF integration or PMS integration based on the input.
    This function supports broadcasting where x_array has shape (m,) and
    distribution has shape (n, m).

    Args:
    -----
        distribution: The distribution array to integrate.
            It should have a shape of (n, m).
        x_array: The x-values array for PDF
            integration. It should have a shape of (m,).
            If None, PMS integration is performed. Defaults to None.
        axis: The axis along which to perform the integration
            for PDF or the sum for PMS.
            Defaults to 0.

    Return:
    -------
        np.ndarray: The result of the integration. If PDF integration is
        performed, the result will have a shape of (n,) if axis=0 or (m,)
        if axis=1. If PMS integration is performed, the result will be a
        single value if axis=None, or an array with reduced dimensionality
        otherwise.
    """
    if x_array is not None:
        # Perform PDF integration along the specified axis
        return np.trapz(y=distribution, x=x_array, axis=axis)
    # Perform PMS integration
    return np.sum(distribution, axis=axis)



""" debye function
"""
import numpy as np

from particula import u


def df1(var):
    """ debye function
    """
    if isinstance(var, u.Quantity):
        var = var.m
    xvar = np.linspace(0, var, 1000)
    return np.trapz(
        xvar[1:]/(np.exp(xvar[1:])-1),
        xvar[1:],
        axis=0
    )/var



"""Callable to generate settings file from template."""
# pylint: disable=dangerous-default-value, too-many-locals, too-many-arguments


from typing import List, Optional
import json
import warnings
import os

from particula.data.loader import get_files_in_folder_with_size


def for_general_1d_load(
    relative_data_folder: str = 'instrument_data',
    filename_regex: str = '*.csv',
    file_min_size_bytes: int = 10,
    header_row: int = 0,
    data_checks: Optional[dict] = None,
    data_column: list = [3, 5],
    data_header: List[str] = ['data 1', 'data 3'],
    time_column: List[int] = [0, 1],
    time_format: str = '%Y-%m-%d %H:%M:%S.%f',
    delimiter: str = ',',
    time_shift_seconds: int = 0,
    timezone_identifier: str = 'UTC',
) -> dict:
    """
    Generate a settings dictionary for loading and checking 1D data from CSV
    files.

    Parameters:
    ------------
    - relative_data_folder (str): The folder path relative to the main script
        where data files are located. Default is 'instrument_data'.
    - filename_regex (str): Regular expression pattern to match filenames in
        the data folder. Default is '*.csv'.
    - file_min_size_bytes (int): Minimum size in bytes for files to be
        considered valid. Default is 10.
    - header_row (int): The index of the row containing column headers
        (0-indexed). Default is 0.
    - data_checks (Optional[dict]): A dictionary containing data quality
        checks such as character length, required character counts, rows to
        skip at the beginning or end. Defaults to basic checks if None.
    - data_column (list of int): List of indices for columns containing data
        points to be loaded. Default is [3, 5].
    - data_header (List[str]): List of strings representing the header names
        for data columns. Default is ['data 1', 'data 3'].
    - time_column (List[int]): List of indices for columns containing time
        information. Default is [0, 1].
    - time_format (str): String format for parsing time columns, using
        strftime conventions. Default is '%Y-%m-%d %H:%M:%S.%f'.
    - delimiter (str): Character used to separate values in the file.
        Default is ','.
    - time_shift_seconds (int): Number of seconds by which to shift time data
        (positive or negative). Default is 0.
    - timezone_identifier (str): Timezone identifier for time conversion.
        Default is 'UTC'.

    Returns:
    - dict: A dictionary with settings for data loading procedures including
        file paths, size requirements, header information, and data check
        parameters.
    """
    if data_checks is None:
        data_checks = {
            "characters": [10, 100],
            "char_counts": {",": 4, ":": 0},
            "skip_rows": 0,
            "skip_end": 0,
        }
    return {
        'relative_data_folder': relative_data_folder,
        'filename_regex': filename_regex,
        'MIN_SIZE_BYTES': file_min_size_bytes,
        'data_loading_function': 'general_1d_load',
        'header_row': header_row,
        'data_checks': data_checks,
        'data_column': data_column,
        'data_header': data_header,
        'time_column': time_column,
        'time_format': time_format,
        'delimiter': delimiter,
        'time_shift_seconds': time_shift_seconds,
        'timezone_identifier': timezone_identifier,
    }


def for_general_sizer_1d_2d_load(
        relative_data_folder: str = 'instrument_data',
        filename_regex: str = '*.csv',
        file_min_size_bytes: int = 10,
        header_row: int = 0,
        data_checks: Optional[dict] = None,
        data_1d_column: list = [3, 5],
        data_1d_header: List[str] = ['data 1', 'data 3'],
        data_2d_dp_start_keyword: str = "Date Time",
        data_2d_dp_end_keyword: str = "Total Conc",
        data_2d_convert_concentration_from: str = "dw/dlogdp",  # or dw
        time_column: List[int] = [0, 1],
        time_format: str = '%Y-%m-%d %H:%M:%S.%f',
        delimiter: str = ',',
        time_shift_seconds: int = 0,
        timezone_identifier: str = 'UTC',
) -> tuple:
    """
    Generate settings for the 1D general file loader and the 2D general sizer
        file loader.

    Parameters:
    - relative_data_folder (str): Path to the folder containing data files,
        relative to the script's location.
    - filename_regex (str): Regex pattern to match filenames for loading.
    - file_min_size_bytes (int): Minimum file size in bytes for a file to be
        considered valid for loading.
    - header_row (int): Row index for the header (0-based) in the data files.
    - data_checks (dict, optional): Specifications for data integrity checks
        to apply when loading data.
    - data_1d_column (list of int): Column indices for 1D data extraction.
    - data_1d_header (list of str): Header names corresponding to the
        `data_1d_column` indices.
    - data_2d_dp_start_keyword (str): Keyword indicating the start of 2D data
        points in a file.
    - data_2d_dp_end_keyword (str): Keyword indicating the end of 2D data
        points in a file.
    - data_2d_convert_concentration_from (str, optional): Unit to convert from
        if concentration scaling is needed for 2D data.
    - time_column (list of int): Column indices for time data extraction.
    - time_format (str): Format string for parsing time data.
    - delimiter (str): Delimiter character for splitting data in the file.
    - time_shift_seconds (int): Seconds to shift the time data by.
    - timezone_identifier (str): Timezone ID for time data interpretation.

    Returns:
    - tuple of (dict, dict): A tuple containing two dictionaries with settings
        for the 1D and 2D data loaders.

    The function defaults `data_checks` to basic validation criteria if not
        provided. It returns separate dictionaries for settings applicable to
        1D and 2D data loaders, which include file paths, size checks, and
        data parsing rules.
    """
    if data_checks is None:
        data_checks = {
            "characters": [10, 100],
            "char_counts": {",": 4, ":": 0},
            "skip_rows": 0,
            "skip_end": 0,
        }
    settings_1d = {
        'relative_data_folder': relative_data_folder,
        'filename_regex': filename_regex,
        'MIN_SIZE_BYTES': file_min_size_bytes,
        'data_loading_function': 'general_1d_load',
        'header_row': header_row,
        'data_checks': data_checks,
        'data_column': data_1d_column,
        'data_header': data_1d_header,
        'time_column': time_column,
        'time_format': time_format,
        'delimiter': delimiter,
        'time_shift_seconds': time_shift_seconds,
        'timezone_identifier': timezone_identifier,
    }
    settings_2d = {
        'relative_data_folder': relative_data_folder,
        'filename_regex': filename_regex,
        'MIN_SIZE_BYTES': file_min_size_bytes,
        'data_loading_function': 'general_2d_load',
        'header_row': header_row,
        'data_checks': data_checks,
        'data_sizer_reader': {
            "Dp_start_keyword": data_2d_dp_start_keyword,
            "Dp_end_keyword": data_2d_dp_end_keyword,
            "convert_scale_from": data_2d_convert_concentration_from,
        },
        'time_column': time_column,
        'time_format': time_format,
        'delimiter': delimiter,
        'time_shift_seconds': time_shift_seconds,
        'timezone_identifier': timezone_identifier,
    }
    return settings_1d, settings_2d


def load_settings_for_stream(
    path: str,
    subfolder: str,
    settings_suffix: str = ''
) -> dict:
    """
    Load settings for Stream data from a JSON file.

    Given a path and subfolder, this function searches for a JSON file
    named 'stream_settings' with an optional suffix. It returns the settings
    as a dictionary. If no file is found, or multiple files are found,
    appropriate errors or warnings are raised.

    Args:
    - path: The path where the subfolder is located.
    - subfolder: The subfolder where the settings file is expected.
    - settings_suffix: An optional suffix for the settings
        file name. Default is an empty string.

    Returns:
    - dict: A dictionary of settings loaded from the file.

    Raises:
    - FileNotFoundError: If no settings file is found.
    - Warning: If more than one settings file is found.
    """
    settings_file_name = f'stream_settings{settings_suffix}.json'
    _, full_path, file_size_in_bytes = get_files_in_folder_with_size(
        path=path,
        subfolder=subfolder,
        filename_regex=settings_file_name,
        min_size=10)

    if len(file_size_in_bytes) == 0:
        raise FileNotFoundError(
            f'No stream_settings file found in {path}/{subfolder}.')
    if len(file_size_in_bytes) > 1:
        warnings.warn(
            f'More than one stream_settings file found in {path}/{subfolder}. '
            'Using the first one found.')

    with open(full_path[0], 'r', encoding='utf-8') as file:
        return json.load(file)


def save_settings_for_stream(
    settings: dict,
    path: str,
    subfolder: str,
    settings_suffix: str = ''
) -> None:
    """
    Save settings for lake data to a JSON file.

    Given a dictionary of settings, this function saves it to a JSON file
    named 'stream_settings' with an optional suffix in the specified filename.
    The JSON file is formatted with a 4-space indentation.

    Args:
    - settings: The settings dictionary to be saved.
    - path: The path where the subfolder is located.
    - subfolder: The subfolder where the settings file will be saved.
    - settings_suffix: An optional suffix for the settings
        file name. Default is an empty string.

    Returns:
    - None
    """
    settings_file_name = f'stream_settings{settings_suffix}.json'
    save_path = os.path.join(path, subfolder, settings_file_name)

    with open(save_path, 'w', encoding='utf-8') as file:
        json.dump(settings, file, indent=4)
        file.write('\n')  # Add an empty line at the bottom


def load_settings_for_lake(
    path: str,
    subfolder: str = '',
    settings_suffix: str = ''
) -> dict:
    """
    Load settings for Lake data from a JSON file. The settings file is
    a dictionary of stream settings dictionaries.

    Given a path and subfolder, this function searches for a JSON file
    named 'lake_settings' with an optional suffix. It returns the settings
    as a dictionary. If no file is found, or multiple files are found,
    appropriate errors or warnings are raised.

    Args:
    - path: The path where the subfolder is located.
    - subfolder: The subfolder where the settings file is expected.
    - settings_suffix: An optional suffix for the settings
        file name. Default is an empty string.

    Returns:
    - dict: A dictionary of settings loaded from the file.

    Raises:
    - FileNotFoundError: If no settings file is found.
    - Warning: If more than one settings file is found.
    """
    settings_file_name = f'lake_settings{settings_suffix}.json'
    _, full_path, file_size_in_bytes = get_files_in_folder_with_size(
        path=path,
        subfolder=subfolder,
        filename_regex=settings_file_name,
        min_size=10)

    if len(file_size_in_bytes) == 0:
        raise FileNotFoundError(
            f'No lake_settings file found in {path}/{subfolder}.')
    if len(file_size_in_bytes) > 1:
        warnings.warn(
            f'More than one lake_settings file found in {path}/{subfolder}. '
            'Using the first one found.')

    with open(full_path[0], 'r', encoding='utf-8') as file:
        return json.load(file)


def save_settings_for_lake(
    settings: dict,
    path: str,
    subfolder: str = '',
    settings_suffix: str = ''
) -> None:
    """
    Save settings for lake data to a JSON file.

    Given a dictionary of settings, this function saves it to a JSON file
    named 'lake_settings' with an optional suffix in the specified filename.
    The JSON file is formatted with a 4-space indentation.

    Args:
    - settings: The settings dictionary to be saved.
    - path: The path where the subfolder is located.
    - subfolder: The subfolder where the settings file will be saved.
    - settings_suffix: An optional suffix for the settings
        file name. Default is an empty string.

    Returns:
    - None
    """
    settings_file_name = f'lake_settings{settings_suffix}.json'
    save_path = os.path.join(path, subfolder, settings_file_name)

    with open(save_path, 'w', encoding='utf-8') as file:
        json.dump(settings, file, indent=4)
        file.write('\n')  # Add an empty line at the bottom



""" testing the coagulation rate
"""
import numpy as np
import pytest
from particula import u
from particula.particle import \
    Particle
from particula.util.coagulation_rate import CoagulationRate
from particula.util.dimensionless_coagulation import full_coag

distribution = Particle(
    cutoff=.9999,
    mode=100e-9,
    nbins=1000,
    nparticles=1e5,
    gsigma=1.25,
).particle_distribution()


fine_distribution = Particle(
    cutoff=.9999,
    mode=100e-9,
    nbins=3000,
    nparticles=1e5,
    gsigma=1.25,
).particle_distribution()


radius = Particle(
    cutoff=.9999,
    mode=100e-9,
    nbins=1000,
    nparticles=1e5,
    gsigma=1.25,
).particle_radius

fine_radius = Particle(
    cutoff=.9999,
    mode=100e-9,
    nbins=3000,
    nparticles=1e5,
    gsigma=1.25,
).particle_radius


kernel = full_coag(
    radius=radius,
    mode=100e-9,
    nbins=1000,
    nparticles=1e5,
    gsigma=1.25,
)

fine_kernel = full_coag(
    radius=fine_radius,
    mode=100e-9,
    nbins=3000,
    nparticles=1e5,
    gsigma=1.25,
)

CoagRate = CoagulationRate(
    distribution=distribution,
    radius=radius,
    kernel=kernel,
)

fine_CoagRate = CoagulationRate(
    distribution=fine_distribution,
    radius=fine_radius,
    kernel=fine_kernel,
)

rads = radius
fine_rads = fine_radius

lnds = distribution
fine_lnds = fine_distribution

kern = kernel
fine_kern = fine_kernel

loss = CoagRate.coag_loss()
fine_loss = fine_CoagRate.coag_loss()

gain = CoagRate.coag_gain()
fine_gain = fine_CoagRate.coag_gain()


def test_kern():
    """ first test the kernel
    """

    assert kern.u == u.m**3/u.s
    assert kern.m.shape == rads.shape + rads.shape
    assert fine_kern.m.shape == fine_rads.shape + fine_rads.shape


def test_loss():
    """ test the loss
    """

    assert loss.u == u.m**-4/u.s
    assert loss.m.shape == rads.shape
    assert loss.m.shape == lnds.shape
    assert fine_loss.m.shape == fine_rads.shape
    assert fine_loss.m.shape == fine_lnds.shape


def test_gain():
    """ test the gain
    """

    assert gain.size == rads.size
    assert fine_gain.size == fine_rads.size
    assert gain.u == u.m**-4/u.s


def test_mass():
    """ test mass conservation
    """

    assert np.trapz((gain - loss)*rads**3, rads).u == u.s**-1
    assert np.trapz((gain - loss)*rads**3, rads) <= 0.2 * u.s**-1
    assert np.trapz(
        (gain - loss).to(u.cm**-4/u.s)*rads**3, rads
    ).m == pytest.approx(0.0)

    assert (
        np.absolute(np.trapz((gain - loss)*rads**3, rads))
        <=
        0.005*np.trapz(gain*rads**3, rads)
    )

    assert (
        np.absolute(np.trapz((gain - loss)*rads**3, rads))
        <=
        0.005*np.trapz(loss*rads**3, rads)
    )


def test_rads():
    """ test radii
    """

    assert rads.u == u.m


def test_res():
    """ testing that mass conservation improves with resolution
    """

    assert (
        np.absolute(np.trapz((gain - loss)*rads**3, rads))
        >=
        np.absolute(np.trapz((fine_gain - fine_loss)*fine_rads**3, fine_rads))
    )
